from typing import Optional, Any
from .definitions import (DocumentType)
from . import locator, match_chain, scr_context
import lxml.html
import urllib


class Document:
    document_type: DocumentType
    path: str
    path_parsed: urllib.parse.ParseResult
    encoding: Optional[str]
    forced_encoding: bool
    text: Optional[str]
    xml: Optional[lxml.html.HtmlElement]
    src_mc: Optional['match_chain.MatchChain']
    locator_match: Optional['locator.LocatorMatch']
    dfmatch: Optional[str]

    def __init__(
        self, document_type: DocumentType, path: str,
        src_mc: Optional['match_chain.MatchChain'],
        match_chains: Optional[list['match_chain.MatchChain']] = None,
        expand_match_chains_above: Optional[int] = None,
        locator_match: Optional['locator.LocatorMatch'] = None,
        path_parsed: Optional[urllib.parse.ParseResult] = None
    ) -> None:
        self.document_type = document_type
        self.path = path
        if path_parsed is not None:
            self.path_parsed = path_parsed
        else:
            self.path_parsed = urllib.parse.urlparse(path)
        self.encoding = None
        self.forced_encoding = False
        self.text = None
        self.xml = None
        self.src_mc = src_mc
        self.locator_match = locator_match
        self.dfmatch = None
        if not match_chains:
            self.match_chains = []
        else:
            self.match_chains = sorted(
                match_chains, key=lambda mc: mc.chain_id)
        self.expand_match_chains_above = expand_match_chains_above

    def __key__(self) -> tuple[DocumentType, str]:
        return (self.document_type, self.path)

    def __eq__(self, other: Any) -> bool:
        return isinstance(self, other.__class__) and self.__key__() == other.__key__()

    def __hash__(self) -> int:
        return hash(self.__key__())

    def decide_encoding(self, ctx: 'scr_context.ScrContext') -> str:
        forced = False
        mc = self.src_mc
        if not mc:
            mc = ctx.match_chains[0]
        if mc.forced_document_encoding:
            enc = mc.forced_document_encoding
            forced = True
        elif self.encoding:
            enc = self.encoding
        else:
            enc = mc.default_document_encoding
        self.encoding = enc
        self.forced_encoding = forced
        return enc



from typing import Any, Optional, Callable


class ConfigDataClass:
    _config_slots_: list[str] = []
    _subconfig_slots_: list[str] = []
    _final_values_: set[str]
    _value_sources_: dict[str, str]

    def __init__(self, blank: bool = False) -> None:
        self._final_values_ = set()
        self._value_sources_ = {}
        if not blank:
            return
        for k in self.__class__._config_slots_:
            self.__dict__[k] = None

    @staticmethod
    def _previous_annotations_as_config_slots(
        annotations: dict[str, Any],
        subconfig_slots: list[str]
    ) -> list[str]:
        subconfig_slots_dict = set(subconfig_slots + ["__annotations__"])
        return list(k for k in annotations.keys() if k not in subconfig_slots_dict)

    def apply_defaults(self, defaults: 'ConfigDataClass') -> None:
        for cs in self.__class__._config_slots_:
            if cs in defaults.__dict__:
                def_val = defaults.__dict__[cs]
            else:
                def_val = defaults.__class__.__dict__[cs]
            if cs not in self.__dict__ or self.__dict__[cs] is None:
                self.__dict__[cs] = def_val
                vs = defaults._value_sources_.get(cs, None)
                if vs:
                    self._value_sources_[cs] = vs

        for scs in self.__class__._subconfig_slots_:
            self.__dict__[scs].apply_defaults(defaults.__dict__[scs])

    def follow_attrib_path(self, attrib_path: list[str]) -> tuple['ConfigDataClass', str]:
        assert(len(attrib_path))
        conf = self
        for attr in attrib_path[:-1]:
            assert attr in conf._subconfig_slots_
            conf = conf.__dict__[attr]
        attr = attrib_path[-1]
        assert attr in conf._config_slots_
        return conf, attr

    def resolve_attrib_path(
        self, attrib_path: list[str],
        transform: Optional[Callable[[Any], Any]] = None
    ) -> Any:
        conf, attr = self.follow_attrib_path(attrib_path)
        if attr in conf.__dict__:
            val = conf.__dict__[attr]
            if transform:
                val = transform(val)
                conf.__dict__[attr] = val
            return val
        val = conf.__class__.__dict__[attr]
        if transform:
            val = transform(val)
            conf.__class__.__dict__[attr] = val
        return val

    def has_custom_value(self, attrib_path: list[str]) -> bool:
        conf, attr = self.follow_attrib_path(attrib_path)
        return attr in conf._value_sources_

    def get_configuring_argument(self, attrib_path: list[str]) -> Optional[str]:
        conf, attr = self.follow_attrib_path(attrib_path)
        return conf._value_sources_.get(attr, None)

    def try_set_config_option(self, attrib_path: list[str], value: Any, arg: str) -> Optional[str]:
        conf, attr = self.follow_attrib_path(attrib_path)
        if attr in conf._final_values_:
            return conf._value_sources_[attr]
        conf._final_values_.add(attr)
        conf._value_sources_[attr] = arg
        conf.__dict__[attr] = value
        return None



from .definitions import (ScrSetupError, ScrMatchError, Verbosity)
from . import match_chain, scr, utils, document, content_match, selenium_setup
from .config_data_class import ConfigDataClass
from typing import Optional, Union, Any, cast
import lxml.etree
import lxml.html
import re
from selenium.common.exceptions import WebDriverException as SeleniumWebDriverException
from selenium.common.exceptions import JavascriptException as SeleniumJavascriptException
from selenium.webdriver.remote.webdriver import WebDriver as SeleniumWebDriver
from urllib3.exceptions import MaxRetryError as SeleniumMaxRetryError
import textwrap


class LocatorMatch:
    xmatch: Optional[str] = None
    xmatch_xml: Optional[lxml.html.HtmlElement] = None
    rmatch: Optional[str] = None
    fres: Optional[str] = None
    jsres: Optional[str] = None
    result: str = ""
    named_cgroups: Optional[dict[str, str]] = None
    unnamed_cgroups: Optional[list[str]] = None

    def set_regex_match(self, match: re.Match[str]) -> None:
        self.result = match.group(0)
        self.rmatch = self.result
        self.named_cgroups = {
            k: (v if v is not None else "")
            for (k, v) in match.groupdict().items()
        }
        self.unnamed_cgroups = [
            x if x is not None else "" for x in match.groups()
        ]

    def __key__(self) -> tuple[Optional[str], Optional[str], Optional[str], Optional[str]]:
        # we only ever compare locator matches from the same match chain
        # therefore it is enough that the complete match is equivalent
        return (self.xmatch, self.rmatch, self.fres, self.jsres)

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, self.__class__) and self.__key__() == other.__key__()

    def __hash__(self) -> int:
        return hash(self.__key__())

    def unnamed_group_list_to_dict(self, name_prefix: str) -> dict[str, str]:
        if self.rmatch is None:
            return {}
        group_dict = {f"{name_prefix}0": self.rmatch}
        for i, g in enumerate(cast(list[str], self.unnamed_cgroups)):
            group_dict[f"{name_prefix}{i+1}"] = g
        return group_dict

    def clone(self) -> 'LocatorMatch':
        c = LocatorMatch()
        if self.xmatch is not None:
            c.xmatch = self.xmatch
        if self.xmatch_xml is not None:
            c.xmatch_xml = self.xmatch_xml
        if self.rmatch is not None:
            c.rmatch = self.rmatch
        if self.fres is not None:
            c.fres = self.fres
        if self.jsres is not None:
            c.jsres = self.jsres
        if self.result is not None:
            c.result = self.result
        if self.named_cgroups is not None:
            c.named_cgroups = self.named_cgroups
        if self.unnamed_cgroups is not None:
            c.unnamed_cgroups = self.unnamed_cgroups
        return c


class Locator(ConfigDataClass):
    name: str
    xpath: Optional[Union[str, lxml.etree.XPath]] = None
    regex: Optional[Union[str, re.Pattern[str]]] = None
    js_script: Optional[str] = None
    format: Optional[str] = None
    multimatch: bool = True
    interactive: bool = False
    __annotations__: dict[str, type]

    _config_slots_: list[str] = (
        ConfigDataClass._previous_annotations_as_config_slots(
            __annotations__, []
        )
    )

    validated: bool = False

    def __init__(self, name: str, blank: bool = False) -> None:
        super().__init__(blank)
        self.name = name

    def is_active(self) -> bool:
        return any(x is not None for x in [self.xpath, self.regex, self.format, self.js_script])

    def parses_documents(self) -> bool:
        return any(x is not None for x in [self.xpath, self.regex, self.js_script])

    def setup_xpath(self, mc: 'match_chain.MatchChain') -> None:
        if self.xpath is None:
            return
        try:
            xp = lxml.etree.XPath(cast(str, self.xpath))
            xp.evaluate(  # type: ignore
                lxml.html.fromstring("<div>test</div>")
            )
        except (lxml.etree.XPathError):
            # don't use the XPathSyntaxError message because they are spectacularily bad
            # e.g. XPath("/div/text(") -> XPathSyntaxError("Missing closing CURLY BRACE")
            raise ScrSetupError(
                f"invalid xpath in {self.get_configuring_argument(['xpath'])}"
            )
        self.xpath = xp

    def gen_dummy_locator_match(self) -> LocatorMatch:
        lm = LocatorMatch()
        if self.xpath:
            lm.xmatch = ""
        if self.regex and type(self.regex) is re.Pattern:
            lm.rmatch = ""
            capture_group_keys = list(self.regex.groupindex.keys())
            unnamed_regex_group_count = (
                self.regex.groups - len(capture_group_keys)
            )
            lm.named_cgroups = {k: "" for k in capture_group_keys}
            lm.unnamed_cgroups = [""] * unnamed_regex_group_count
        if self.format:
            lm.fres = ""
        if self.js_script:
            lm.jsres = ""
        return lm

    def setup_regex(self, mc: 'match_chain.MatchChain') -> None:
        if self.regex is None:
            return
        try:
            self.regex = re.compile(self.regex, re.DOTALL | re.MULTILINE)
        except re.error as err:
            raise ScrSetupError(
                f"invalid regex ({err.msg}) in {self.get_configuring_argument(['regex'])}"
            )

    def setup_format(self, mc: 'match_chain.MatchChain') -> None:
        if self.format is None:
            return
        scr.validate_format(
            self, ["format"], mc.gen_dummy_content_match(), True, False
        )

    def setup_js(self, mc: 'match_chain.MatchChain') -> None:
        if self.js_script is None:
            return
        args_dict: dict[str, Any] = {}
        dummy_doc = mc.gen_dummy_document()
        scr.apply_general_format_args(dummy_doc, mc, args_dict, ci=None)
        scr.apply_locator_match_format_args(
            self.name, self.gen_dummy_locator_match(), args_dict
        )
        js_prelude = ""
        for i, k in enumerate(args_dict.keys()):
            js_prelude += f"const {k} = arguments[{i}];\n"
        self.js_script = js_prelude + self.js_script

    def setup(self, mc: 'match_chain.MatchChain') -> None:
        self.xpath = utils.empty_string_to_none(cast(str, self.xpath))
        assert self.regex is None or type(self.regex) is str
        self.regex = utils.empty_string_to_none(self.regex)
        self.format = utils.empty_string_to_none(self.format)
        self.setup_xpath(mc)
        self.setup_regex(mc)
        self.setup_js(mc)
        self.setup_format(mc)
        self.validated = True

    def match_xpath(
        self,
        src_text: str,
        src_xml: Optional[lxml.html.HtmlElement],
        doc_path: str,
        store_xml: bool = False
    ) -> list[LocatorMatch]:
        if self.xpath is None:
            lm = LocatorMatch()
            lm.result = src_text
            return [lm]
        assert src_xml is not None
        try:
            xp = cast(lxml.etree.XPath, self.xpath)
            if type(src_xml) == lxml.etree._ElementUnicodeResult:  # type: ignore
                # since lxml doesn't allow us to evaluate xpaths on these,
                # but we need it for lic, we hack in support for it by
                # generating a derived xpath that gets the expected results while
                # actually being evaluated on the parent
                fixed_xpath = f"./@{src_xml.attrname}"
                if xp.path[0:1] != "/":
                    fixed_xpath += "/"
                fixed_xpath += xp.path
                xpath_matches = src_xml.getparent().xpath(fixed_xpath)
            else:
                xpath_matches = (xp.evaluate(src_xml))  # type: ignore
        except lxml.etree.XPathError:
            raise ScrMatchError(
                f"xpath matching failed for: '{self.xpath}' in {doc_path}"
            )
        except lxml.etree.LxmlError as ex:
            raise ScrMatchError(
                f"xpath '{self.xpath}' to {doc_path}: "
                + f"{ex.__class__.__name__}:  {str(ex)}"
            )

        if not isinstance(xpath_matches, list):
            raise ScrMatchError(
                f"xpath matching failed for: '{self.xpath}' in {doc_path}"
            )

        if len(xpath_matches) > 1 and not self.multimatch:
            xpath_matches = xpath_matches[:1]
        res = []
        for xm in xpath_matches:
            lm = LocatorMatch()
            if type(xm) == lxml.etree._ElementUnicodeResult:  # type: ignore
                lm.xmatch = str(xm)
                if store_xml:
                    lm.xmatch_xml = xm
            else:
                try:
                    lm.result = lxml.html.tostring(xm, encoding="unicode")
                    lm.xmatch = lm.result
                    if store_xml:
                        lm.xmatch_xml = xm
                except (lxml.etree.LxmlError, UnicodeEncodeError) as ex1:
                    raise ScrMatchError(
                        f"{doc_path}: xpath match encoding failed: {str(ex1)}"
                    )
            lm.result = lm.xmatch
            res.append(lm)
        return res

    def apply_regex_matches(
        self, lms: list[LocatorMatch],
        multimatch: Optional[bool] = None
    ) -> list[LocatorMatch]:
        if self.regex is None:
            return lms
        rgx = cast(re.Pattern[str], self.regex)
        if multimatch is None:
            multimatch = self.multimatch

        lms_new = []
        for lm in lms:
            if not multimatch:
                match = rgx.match(lm.result)
                if match:
                    lm.set_regex_match(match)
                    lms_new.append(lm)
                continue
            res: Optional[LocatorMatch] = lm
            for match in rgx.finditer(lm.result):
                if res is None:
                    res = lm.clone()
                res.set_regex_match(match)
                lms_new.append(res)
                if not multimatch:
                    break
                res = None
        return lms_new

    def apply_js_matches(
        self, doc: 'document.Document', mc: 'match_chain.MatchChain', lms: list['LocatorMatch'],
        multimatch: Optional[bool] = None
    ) -> list['LocatorMatch']:
        if self.js_script is None:
            return lms
        if multimatch is None:
            multimatch = self.multimatch
        lms_new: list[LocatorMatch] = []
        for lm in lms:
            args_dict: dict[str, Any] = {}
            scr.apply_general_format_args(doc, mc, args_dict, ci=None)
            scr.apply_locator_match_format_args(self.name, lm, args_dict)
            try:
                mc.js_executed = True
                drv = cast(SeleniumWebDriver, mc.ctx.selenium_driver)

                results = drv.execute_script(
                    self.js_script, *args_dict.values())  # type: ignore

            except SeleniumJavascriptException as ex:
                arg = cast(str, self.get_configuring_argument(['js_script']))
                name = arg[0: arg.find("=")]
                scr.log(
                    mc.ctx, Verbosity.WARN,
                    f"{name}: js exception on {utils.truncate(doc.path)}:\n{textwrap.indent(str(ex), '    ')}"
                )
                continue
            except (SeleniumWebDriverException, SeleniumMaxRetryError):
                if selenium_setup.selenium_has_died(mc.ctx):
                    raise ScrMatchError(
                        "the selenium instance was closed unexpectedly")
                continue
            if results is None:
                continue
            if type(results) is not list:
                results = [str(results)]
            res: Optional[LocatorMatch] = lm
            for r in results:
                if res is None:
                    res = lm.clone()
                res.jsres = r
                res.result = r
                lms_new.append(res)
                if not multimatch:
                    break
                res = None
        return lms_new

    def apply_format_for_content_match(
        self, cm: 'content_match.ContentMatch', lm: LocatorMatch
    ) -> None:
        if not self.format:
            return
        lm.fres = self.format.format(**scr.content_match_build_format_args(cm))
        lm.result = lm.fres

    def apply_format_for_document_match(
        self, doc: 'document.Document', mc: 'match_chain.MatchChain', lm: 'LocatorMatch'
    ) -> None:
        if not self.format:
            return
        args_dict: dict[str, Any] = {}
        scr.apply_general_format_args(doc, mc, args_dict, ci=None)
        scr.apply_locator_match_format_args(self.name, lm, args_dict)
        lm.fres = self.format.format(**args_dict)
        lm.result = lm.fres

    def is_unset(self) -> bool:
        return min([v is None for v in [self.xpath, self.regex, self.format]])



from typing import TypeVar
from enum import Enum, IntEnum

T = TypeVar("T")
K = TypeVar("K")
V = TypeVar("V")

# because for python, sys.argv[0] does not reflect what the user typed anyways,
# we just use this fixed value for --help etc.
SCRIPT_NAME = "scr"
VERSION = "0.10.1"

SCR_USER_AGENT = f"{SCRIPT_NAME}/{VERSION}"
FALLBACK_DOCUMENT_SCHEME = "https"
DEFAULT_TIMEOUT_SECONDS = 30
URL_FILENAME_MAX_LEN = 256


# mimetype to use for selenium downloading to avoid triggering pdf viewers etc.
DUMMY_MIMETYPE = "application/zip"

DEFAULT_CPF = "{c}\\n"
DEFAULT_CWF = "{c}"
DEFAULT_CSF = "{fn}"
DEFAULT_ESCAPE_SEQUENCE = "<END>"


class ScrSetupError(Exception):
    pass


class ScrFetchError(Exception):
    pass


class ScrMatchError(Exception):
    pass


class InteractiveResult(Enum):
    ACCEPT = 0
    REJECT = 1
    EDIT = 2
    INSPECT = 3
    SKIP_CHAIN = 4
    SKIP_DOC = 5
    ACCEPT_CHAIN = 6
    ERROR = 0


class SeleniumVariant(Enum):
    DISABLED = 0
    CHROME = 1
    FIREFOX = 2
    TORBROWSER = 3

    @staticmethod
    def default() -> 'SeleniumVariant':
        return SeleniumVariant.FIREFOX

    def enabled(self) -> bool:
        return self != SeleniumVariant.DISABLED


class SeleniumDownloadStrategy(Enum):
    EXTERNAL = 0
    INTERNAL = 1
    FETCH = 2


class SeleniumStrategy(Enum):
    DISABLED = 0
    PLAIN = 1
    ANYMATCH = 2
    INTERACTIVE = 3
    DEDUP = 4


class Verbosity(IntEnum):
    ERROR = 1
    WARN = 2
    INFO = 3
    DEBUG = 4


class DocumentType(Enum):
    URL = 1
    FILE = 2
    RFILE = 3
    CONTENT_MATCH = 4
    CONTENT_FILE = 999  # special type for the document as content optimization

    def derived_type(self) -> 'DocumentType':
        if self == DocumentType.RFILE:
            return DocumentType.URL
        if self == DocumentType.CONTENT_FILE:
            return DocumentType.FILE
        return self

    def url_handling_type(self) -> 'DocumentType':
        if self == DocumentType.RFILE:
            return DocumentType.FILE
        return self


document_type_display_dict: dict[DocumentType, str] = {
    DocumentType.URL: "url",
    DocumentType.FILE: "file",
    DocumentType.RFILE: "rfile",
    DocumentType.CONTENT_MATCH: "content match from"
}

selenium_variants_dict: dict[str, SeleniumVariant] = {
    "disabled": SeleniumVariant.DISABLED,
    "tor": SeleniumVariant.TORBROWSER,
    "firefox": SeleniumVariant.FIREFOX,
    "chrome": SeleniumVariant.CHROME
}

selenium_variants_display_dict: dict[SeleniumVariant, str] = {
    SeleniumVariant.DISABLED: "disabled",
    SeleniumVariant.TORBROWSER: "Tor Browser",
    SeleniumVariant.FIREFOX: "Firefox",
    SeleniumVariant.CHROME: "Chrome"
}


selenium_download_strategies_dict: dict[str, SeleniumDownloadStrategy] = {
    "external": SeleniumDownloadStrategy.EXTERNAL,
    "internal": SeleniumDownloadStrategy.INTERNAL,
    "fetch": SeleniumDownloadStrategy.FETCH,
}

selenium_strats_dict: dict[str, SeleniumStrategy] = {
    "plain": SeleniumStrategy.PLAIN,
    "anymatch": SeleniumStrategy.ANYMATCH,
    "interactive": SeleniumStrategy.INTERACTIVE,
    "dedup": SeleniumStrategy.DEDUP,
}

verbosities_dict: dict[str, Verbosity] = {
    "error": Verbosity.ERROR,
    "warn": Verbosity.WARN,
    "info": Verbosity.INFO,
    "debug": Verbosity.DEBUG,
}

verbosities_display_dict: dict[Verbosity, str] = {
    Verbosity.ERROR: "[ERROR]: ",
    Verbosity.WARN:  " [WARN]: ",
    Verbosity.INFO:  " [INFO]: ",
    Verbosity.DEBUG: "[DEBUG]: ",
}



from . import windows
from typing import Optional, Callable
import platform
import sys
import re
import select
from .definitions import T


def is_windows() -> bool:
    return platform.system() == 'Windows'


def is_osx() -> bool:
    return platform.system() == 'Darwin'


def is_linux() -> bool:
    return platform.system() == 'Linux'


DEFAULT_TRUNCATION_LENGTH = 200


def truncate(
    text: str,
    max_len: int = DEFAULT_TRUNCATION_LENGTH,
    trailer: str = "..."
) -> str:
    if len(text) > max_len:
        assert(max_len > len(trailer))
        return text[0: max_len - len(trailer)] + trailer
    return text


def begins(string: str, begin: str) -> bool:
    return len(string) >= len(begin) and string[0:len(begin)] == begin


def empty_string_to_none(string: Optional[str]) -> Optional[str]:
    if string == "":
        return None
    return string


def choose_first_not_none(*tries: Callable[[], Optional[T]]) -> Optional[T]:
    for t in tries:
        res = t()
        if res is not None:
            return res
    return None


def stdin_has_content(timeout: float) -> bool:
    assert timeout >= 0
    if sys.platform != 'win32':
        rlist, _, _ = select.select(
            [sys.stdin], [], [], timeout
        )
        return bool(rlist)
    else:
        return windows.stdin_has_content(timeout)


def remove_file_scheme_from_url(url: str) -> str:
    offs = len("file:")
    assert url[:offs] == "file:"
    for i in range(2):
        if url[offs] == "/":
            offs += 1
    url = url[offs:]
    # browsers turn windows paths like 'C:\foobar' into "file:///C:/foobar"
    # which does not fly with pythons os.path, so we hack in a fix
    # that removes that leading slash from the path
    # once we do that, urllib thinks that C: is a scheme,
    # so we have to include file://, see handle_widows_paths normalize_link
    if is_windows() and re.match("/[A-Za-z]:/", url):
        url = url[1:]
    return url



from .definitions import (
    SeleniumVariant, Verbosity, ScrSetupError, selenium_variants_display_dict,
    SCRIPT_NAME
)
from . import scr, utils, scr_context, windows
from typing import Optional, cast
import selenium_driver_updater
import selenium_driver_updater.util.exceptions
import os
import glob
import shutil
from functools import cache
SELENIUM_DRIVER_DIR_ADDED_TO_PATH: bool = False


def try_resolve_executable_names(exec_names: list[str]) -> Optional[str]:
    for x in exec_names:
        path = shutil.which(x)
        if path is not None:
            return path
    return None


def accept_first_existing_path(paths: list[str]) -> Optional[str]:
    for path in paths:
        if os.path.exists(path):
            return path
    return None


def find_chrome_binary() -> Optional[str]:
    if utils.is_linux():
        return try_resolve_executable_names([
            "chromium",
            "chromium-browser",
            "google-chrome-stable",
            "google-chrome"
        ])
    if utils.is_osx():
        return accept_first_existing_path([
            "/Applications/Chromium.app/Contents/MacOS/Chromium",
            '/Applications/Google Chrome.app/Contents/MacOS/Google Chrome',
        ])
    if not utils.is_windows():
        return None

    return utils.choose_first_not_none(
        lambda: windows.try_get_app_path_from_reg_uninstall_path("Chromium", "chrome.exe"),
        lambda: windows.try_get_app_path_from_reg_start_menu_internet("Chromium"),
        lambda: windows.try_get_app_path_from_reg_app_paths("chrome.exe"),
        lambda: windows.try_get_app_path_from_reg_uninstall_path("Google Chrome", "chrome.exe"),
        lambda: windows.try_get_app_path_from_reg_start_menu_internet("Google Chrome"),
    )


@ cache
def get_script_dir() -> str:
    return os.path.abspath(os.path.dirname(__file__))


@ cache
def get_selenium_drivers_dir() -> str:
    script_dir = get_script_dir()
    debug_path = os.path.join(script_dir, "selenium_drivers_debug")
    if os.path.exists(debug_path):
        return debug_path
    return os.path.join(get_script_dir(), "selenium_drivers")


def get_selenium_driver_executable_basename(variant: SeleniumVariant) -> Optional[str]:
    bn = {
        SeleniumVariant.CHROME: "chromedriver",
        SeleniumVariant.FIREFOX: "geckodriver",
        SeleniumVariant.TORBROWSER: "geckodriver",
        SeleniumVariant.DISABLED: None
    }[variant]
    if bn is None:
        return None
    return bn + (".exe" if utils.is_windows() else "")


def get_local_selenium_driver_executable_path(variant: SeleniumVariant) -> Optional[str]:
    basename = get_selenium_driver_executable_basename(variant)
    if basename is None:
        return None
    return os.path.join(
        get_selenium_drivers_dir(),
        basename
    )


def touch_file(path: str) -> None:
    open(path, "ab+").close()


def is_selenium_driver_present(path: str) -> bool:
    try:
        return os.path.getsize(path) != 0
    except FileNotFoundError:
        try:
            touch_file(path)
        except IOError:
            pass
        return False


def try_get_local_selenium_driver_path(variant: 'SeleniumVariant') -> Optional[str]:
    path = get_local_selenium_driver_executable_path(variant)
    if path is None:
        return None
    if is_selenium_driver_present(path):
        return path
    return None


def put_local_selenium_driver_in_path(ctx: 'scr_context.ScrContext', variant: 'SeleniumVariant') -> None:
    global SELENIUM_DRIVER_DIR_ADDED_TO_PATH
    if SELENIUM_DRIVER_DIR_ADDED_TO_PATH:
        return
    SELENIUM_DRIVER_DIR_ADDED_TO_PATH = True
    path_seperator = ":" if not utils.is_windows() else ";"
    os.environ["PATH"] += path_seperator + get_selenium_drivers_dir()
    scr.log(ctx, Verbosity.INFO, f"added to PATH: {get_selenium_drivers_dir()}")


def get_preferred_selenium_driver_path(variant: 'SeleniumVariant') -> str:
    path = try_get_local_selenium_driver_path(variant)
    if path is not None:
        return path
    basename = get_selenium_driver_executable_basename(variant)
    # this function makes no sense for DISABLED
    assert basename is not None
    return basename


def install_selenium_driver(ctx: 'scr_context.ScrContext', variant: 'SeleniumVariant', update: bool) -> None:
    if variant == SeleniumVariant.CHROME:
        driver_name = selenium_driver_updater.DriverUpdater.chromedriver
    elif variant in [SeleniumVariant.FIREFOX, SeleniumVariant.TORBROWSER]:
        driver_name = selenium_driver_updater.DriverUpdater.geckodriver
    else:
        raise ScrSetupError(
            "unable to install webdriver for '{selenium_variants_display_dict[variant]}'"
        )
    driver_dir = get_selenium_drivers_dir()
    local_driver_path = cast(
        str, get_local_selenium_driver_executable_path(variant)
    )
    have_local = is_selenium_driver_present(local_driver_path)
    if have_local and not update:
        scr.log(
            ctx, Verbosity.INFO,
            f"existing {selenium_variants_display_dict[variant]} driver found"
        )
        return
    success = False

    try:
        scr.log(
            ctx, Verbosity.INFO,
            ("updating" if have_local else "installing") +
            f" {selenium_variants_display_dict[variant]} driver ..."
        )
        selenium_driver_updater.DriverUpdater.install(
            path=driver_dir, driver_name=driver_name,
            check_driver_is_up_to_date=have_local,
            enable_library_update_check=False,
            upgrade=True,
            info_messages=False
        )
        success = True
        scr.log(
            ctx, Verbosity.INFO,
            f"{selenium_variants_display_dict[variant]} driver "
            + ("updated" if have_local else "installed")
        )
    except (selenium_driver_updater.util.exceptions.Error) as ex:
        scr.log(
            ctx, Verbosity.ERROR,
            f"failed to fetch {selenium_variants_display_dict[variant]} driver: {str(ex)}"
        )
    finally:
        # we should not have to do this, this is working around a bug in the library
        if not success:
            open(local_driver_path, "w").close()
        # cleanup_selenium_installation_artifacts(ctx, silent=not success)


def cleanup_selenium_installation_artifacts(ctx: 'scr_context.ScrContext', silent: bool = False) -> None:
    # clean up remains from potentially failed installs
    present_files = [
        os.path.abspath(p)
        for p in glob.glob(get_selenium_drivers_dir() + "/**")
    ]
    variant_files = {
        get_local_selenium_driver_executable_path(v)
        for v in SeleniumVariant if v is not None
    }
    cleanup = False
    for pf in present_files:
        if pf not in variant_files:
            os.remove(pf)
            cleanup = True
    if cleanup:
        scr.log(
            ctx, Verbosity.INFO,
            "cleaned up artifacts from previous installations"
        )


def uninstall_selenium_driver(ctx: 'scr_context.ScrContext', variant: 'SeleniumVariant') -> None:
    cleanup_selenium_installation_artifacts(ctx)
    path = try_get_local_selenium_driver_path(variant)
    if path is None:
        scr.log(
            ctx, Verbosity.WARN,
            f"no {selenium_variants_display_dict[variant]} driver (local to {SCRIPT_NAME}) installed"
        )
        return
    open(path, 'w').close()
    scr.log(
        ctx, Verbosity.INFO,
        f"{selenium_variants_display_dict[variant]} driver uninstalled"
    )



from typing import Optional, Any
from . import locator, match_chain, document
import urllib


class ContentMatch:
    clm: 'locator.LocatorMatch'
    llm: Optional['locator.LocatorMatch'] = None
    mc: 'match_chain.MatchChain'
    doc: 'document.Document'

    # these are set once we accept the CM, not during it's creation
    ci: Optional[int] = None
    di: Optional[int] = None

    url_parsed: Optional[urllib.parse.ParseResult]

    def __init__(
        self,
        clm: 'locator.LocatorMatch',
        llm: Optional['locator.LocatorMatch'],
        mc: 'match_chain.MatchChain',
        doc: 'document.Document'
    ) -> None:
        self.llm = llm
        self.clm = clm
        self.mc = mc
        self.doc = doc

    def __key__(self) -> Any:
        return (
            self.doc, self.clm.__key__(),
            self.llm.__key__() if self.llm else None,
        )

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, self.__class__) and other.__key__() == self.__key__()

    def __hash__(self) -> int:
        return hash(self.__key__())



from .definitions import SeleniumVariant, Verbosity, DEFAULT_TIMEOUT_SECONDS
from . import match_chain, download_job, document
from .config_data_class import ConfigDataClass
from typing import Optional, Any
import multiprocessing
import os
from http.cookiejar import MozillaCookieJar
from selenium.webdriver.remote.webdriver import WebDriver as SeleniumWebDriver
from collections import deque


class ScrContext(ConfigDataClass):
    # config members
    cookie_file: Optional[str] = None
    exit: bool = False
    selenium_variant: SeleniumVariant = SeleniumVariant.DISABLED
    selenium_headless: bool = False
    tor_browser_dir: Optional[str] = None
    user_agent_random: Optional[bool] = False
    user_agent: Optional[str] = None
    verbosity: Verbosity = Verbosity.WARN
    documents_bfs: bool = False
    selenium_keep_alive: bool = False
    repl: bool = False
    request_timeout_seconds: float = DEFAULT_TIMEOUT_SECONDS
    max_download_threads: int = multiprocessing.cpu_count()

    selenium_log_path: str = os.devnull
    selenium_poll_frequency_secs: float = 0.3
    selenium_content_count_pad_length: int = 6
    downloads_temp_dir: Optional[str] = None
    download_tmp_index: int = 0

    # not really config, but recycled in repl mode
    dl_manager: Optional['download_job.DownloadManager'] = None
    cookie_jar: Optional[MozillaCookieJar] = None
    cookie_dict: dict[str, dict[str, dict[str, Any]]]
    selenium_driver: Optional[SeleniumWebDriver] = None
    __annotations__: dict[str, type]
    _config_slots_: list[str] = (
        ConfigDataClass._previous_annotations_as_config_slots(
            __annotations__, []
        )
    )

    # non config members
    match_chains: list['match_chain.MatchChain']
    docs: deque['document.Document']
    reused_doc: Optional['document.Document'] = None
    changed_selenium: bool = False
    defaults_mc: 'match_chain.MatchChain'
    origin_mc: 'match_chain.MatchChain'
    error_code: int = 0
    abort: bool = False

    # used for --help --version and selinstall/selupdate to indicate
    # that if there are no match chains etc. we should exit without error
    special_args_occured: bool = False

    def __init__(self, blank: bool = False) -> None:
        super().__init__(blank)
        self.cookie_dict = {}
        self.match_chains = []
        self.docs = deque()
        self.defaults_mc = match_chain.MatchChain(self, -1)
        self.origin_mc = match_chain.MatchChain(self, -1, blank=True)
        # turn ctx to none temporarily for origin so it can be deepcopied
        self.origin_mc.ctx = None  # type: ignore



from typing import Any, Optional, Callable, Iterable
import urllib
import itertools
import copy
from .definitions import (
    T, ScrSetupError, DocumentType, SeleniumVariant, selenium_variants_dict,
    selenium_strats_dict, selenium_download_strategies_dict, verbosities_dict,
    VERSION, SCRIPT_NAME, DEFAULT_ESCAPE_SEQUENCE, DEFAULT_CPF, DEFAULT_CWF,
    DEFAULT_TIMEOUT_SECONDS
)

from . import (
    match_chain, document, input_sequences,
    selenium_driver_download, scr_context, utils, scr
)

import sys
import re

MATCH_CHAIN_ARGUMENT_REGEX = re.compile("^[0-9\\-\\*\\^]*$")


def help(err: bool = False) -> None:
    text = f"""{SCRIPT_NAME} [OPTIONS]

    Matching chains are evaluated in the following order, skipping unspecified steps:
    xpath -> regex -> (javascript) -> python format string

    Content to Write out:
        cx=<xpath>            xpath for content matching
        cr=<regex>            regex for content matching
        cjs=<js string>       javascript to execute on the page, format args are available as js variables (selenium only)
        cf=<format string>    content format string (args: <cr capture groups>, xmatch, rmatch, di, ci)
        cmm=<bool>            allow multiple content matches in one document instead of picking the first (defaults to true)
        cimin=<number>        initial content index, each successful match gets one index
        cimax=<number>        max content index, matching stops here
        cicont=<bool>         don't reset the content index for each document
        csf=<format string>   save content to file at the path resulting from the format string, empty to enable
        cwf=<format string>   format to write to file. defaults to \"{DEFAULT_CWF}\"
        cpf=<format string>   print the result of this format string for each content, empty to disable
                              defaults to \"{DEFAULT_CPF}\" if cpf, csf and cfc are unspecified
        cfc=<chain spec>      forward content match as a virtual document
        cff=<format string>   format of the virtual document forwarded to the cfc chains. defaults to \"{DEFAULT_CWF}\"
        csin<bool>            give a promt to edit the save path for a file
        cin=<bool>            give a prompt to ignore a potential content match
        cl=<bool>             treat content match as a link to the actual content
        cesc=<string>         escape sequence to terminate content in cin mode, defaults to \"{DEFAULT_ESCAPE_SEQUENCE}\"
        cenc=<encoding>       default encoding to assume that content is in
        cfenc=<encoding>      encoding to always assume that content is in, even if http(s) says differently

    Labels to give each matched content (mostly useful for the filename in csf):
        lx=<xpath>           xpath for label matching
        lr=<regex>           regex for label matching
        ljs=<js string>      javascript to execute on the page, format args are available as js variables (selenium only)
        lf=<format string>   label format string
        lic=<bool>           match for the label within the content match instead of the hole document
        las=<bool>           allow slashes in labels
        lmm=<bool>           allow multiple label matches in one document instead of picking the first (for all content matches)
        lam=<bool>           allow missing label (default is to skip content if no label is found)
        lfd=<format string>  default label format string to use if there's no match
        lin=<bool>           give a prompt to edit the generated label

    Further documents to scan referenced in already found ones:
        dx=<xpath>           xpath for document matching
        dr=<regex>           regex for document matching
        djs=<js string>      javascript to execute on the page, format args are available as js variables (selenium only)
        df=<format string>   document format string
        dimin=<number>       initial document index, each successful match gets one index
        dimax=<number>       max document index, matching stops here
        dmm=<bool>           allow multiple document matches in one document instead of picking the first
        din=<bool>           give a prompt to ignore a potential document match
        denc=<encoding>      default document encoding to use for following documents, default is utf-8
        dfenc=<encoding>     force document encoding for following documents, even if http(s) says differently
        dsch=<scheme>        default scheme for urls derived from following documents, defaults to "https"
        dpsch=<bool>         use the parent documents scheme if available, defaults to true unless dsch is specified
        dfsch=<scheme>       force this scheme for urls derived from following documents
        doc=<chain spec>     chains that matched documents should apply to, default is the same chain

    Initial Documents:
        url=<url>            fetch a document from a url, derived document matches are (relative) urls
        file=<path>          fetch a document from a file, derived documents matches are (relative) file pathes
        rfile=<path>         fetch a document from a file, derived documents matches are urls

    Other:
        selstrat=<strategy>  matching strategy for selenium (default: plain, values: anymatch, plain, interactive, deduplicate)
        seldl=<dl strategy>  download strategy for selenium (default: external, values: external, internal, fetch)
        owf=<bool>           allow to overwrite existing files, defaults to true

    Format Args:
        Named arguments for <format string> arguments.
        Some only become available later in the pipeline (e.g. {{cm}} is not available inside cf).

        {{cx}}                 content xpath match
        {{cr}}                 content regex match, equal to {{cx}} if cr is unspecified
        <cr capture groups>  the named regex capture groups (?P<name>...) from cr are available as {{name}},
                             the unnamed ones (...) as {{cg<unnamed capture group number>}}
        {{cf}}                 content after applying cf
        {{cjs}}                output of cjs
        {{cm}}                 final content match after link normalization (cl) and user interaction (cin)
        {{c}}                  content, downloaded from cm in case of cl, otherwise equal to cm

        {{lx}}                 label xpath match
        {{lr}}                 label regex match, equal to {{lx}} if lr is unspecified
        <lr capture groups>  the named regex capture groups (?P<name>...) from cr are available as {{name}},
                             the unnamed ones (...) as {{lg<unnamed capture group number>}}
        {{lf}}                 label after applying lf
        {{ljs}}                output of ljs
        {{l}}                  final label after user interaction (lin)

        {{dx}}                 document link xpath match
        {{dr}}                 document link regex match, equal to {{dx}} if dr is unspecified
        <dr capture groups>  the named regex capture groups (?P<name>...) from dr are available as {{name}},
                             the unnamed ones (...) as {{dg<unnamed capture group number>}}
        {{df}}                 document link after applying df
        {{djs}}                output of djs
        {{d}}                  final document link after user interaction (din)

        {{di}}                 document index
        {{ci}}                 content index
        {{dl}}                 document link (inside df, this refers to the parent document)
        {{cenc}}               content encoding, deduced while respecting cenc and cfenc
        {{cesc}}               escape sequence for separating content, can be overwritten using cesc
        {{chain}}              id of the match chain that generated this content

        {{fn}}                 filename from the url of a cm with cl
        {{fb}}                 basename component of {{fn}} (extension stripped away)
        {{fe}}                 extension component of {{fn}}, including the dot (empty string if there is no extension)


    Chain Syntax:
        Any option above can restrict the matching chains is should apply to using opt<chainspec>=<value>.
        Use "-" for ranges, "," for multiple specifications, and "^" to except the following chains.
        Examples:
            lf1,3-5=foo        sets "lf" to "foo" for chains 1, 3, 4 and 5.
            lf2-^4=bar         sets "lf" to "bar" for all chains larger than or equal to 2, except chain 4

    Miscellaneous:
        help                   prints this help
        selinstall=<browser>   installs selenium driver for the specified browser in the directory of this script
        seluninstall=<browser> uninstalls selenium driver for the specified browser in the directory of this script
        selupdate=<browser>    updates (or installs) the local selenium driver for the specified browser
        version                print version information

    Global Options:
        timeout=<seconds>      seconds before a web request timeouts (default {DEFAULT_TIMEOUT_SECONDS})
        bfs=<bool>             traverse the matched documents in breadth first order instead of depth first
        v=<verbosity>          output verbosity levels (default: warn, values: info, warn, error)
        ua=<string>            user agent to pass in the html header for url GETs
        uar=<bool>             use a rangom user agent
        selkeep=<bool>         keep selenium instance alive after the command finished
        cookiefile=<path>      path to a netscape cookie file. cookies are passed along for url GETs
        sel=<browser|bool>     use selenium (default is firefox) to load urls into an interactive browser session
                               (default: disabled, values: tor, chrome, firefox, disabled)
        selh=<bool>            use selenium in headless mode, implies sel
        tbdir=<path>           root directory of the tor browser installation, implies sel=tor
                               (default: environment variable TOR_BROWSER_DIR)
        mt=<int>               maximum threads for background downloads, 0 to disable. defaults to cpu core count
        repl=<bool>            accept commands in a read eval print loop
        exit=<bool>            exit the repl (with the result of the current command)

        """.strip()
    if err:
        sys.stderr.write(text + "\n")
        sys.exit(1)

    else:
        print(text)


def parse_mc_range_int(ctx: 'scr_context.ScrContext', v: str, arg: str) -> int:
    try:
        return int(v)
    except ValueError:
        raise ScrSetupError(
            f"failed to parse '{v}' as an integer for match chain specification of '{arg}'"
        )


def extend_match_chain_list(ctx: 'scr_context.ScrContext', needed_id: int) -> None:
    if len(ctx.match_chains) > needed_id:
        return
    for i in range(len(ctx.match_chains), needed_id+1):
        mc = copy.deepcopy(ctx.origin_mc)
        mc.chain_id = i
        ctx.match_chains.append(mc)


def parse_simple_mc_range(ctx: 'scr_context.ScrContext', mc_spec: str, arg: str) -> Iterable['match_chain.MatchChain']:
    sections = mc_spec.split(",")
    ranges = []
    for s in sections:
        s = s.strip()
        if s == "":
            raise ScrSetupError(
                "invalid empty range in match chain specification of '{arg}'")
        dash_split = [r.strip() for r in s.split("-")]
        if len(dash_split) > 2 or s == "-":
            raise ScrSetupError(
                "invalid range '{s}' in match chain specification of '{arg}'")
        if len(dash_split) == 1:
            id = parse_mc_range_int(ctx, dash_split[0], arg)
            extend_match_chain_list(ctx, id)
            ranges.append([ctx.match_chains[id]])
        else:
            lhs, rhs = dash_split
            if lhs == "":
                fst = 0
            else:
                fst = parse_mc_range_int(ctx, lhs, arg)
            if rhs == "":
                extend_match_chain_list(ctx, fst)
                snd = len(ctx.match_chains) - 1
                ranges.append([ctx.origin_mc])
            else:
                snd = parse_mc_range_int(ctx, dash_split[1], arg)
                if fst > snd:
                    raise ScrSetupError(
                        f"second value must be larger than first for range {s} "
                        + f"in match chain specification of '{arg}'"
                    )
                extend_match_chain_list(ctx, snd)
            ranges.append(ctx.match_chains[fst: snd + 1])
    return itertools.chain(*ranges)


def match_traditional_cli_arg(arg: str, true_opt_name: str, aliases: set[str]) -> Optional[bool]:
    tolen = len(true_opt_name)
    arglen = len(arg)
    if utils.begins(arg, f"{true_opt_name}"):
        if arglen > tolen:
            if arg[tolen] != "=":
                return None
        return parse_bool_arg(arg[len("{true_opt_name}="):], arg)
    if arg in aliases:
        return True
    return None


def parse_mc_range(ctx: 'scr_context.ScrContext', mc_spec: str, arg: str) -> Iterable['match_chain.MatchChain']:
    if mc_spec == "":
        return [ctx.defaults_mc]

    esc_split = [x.strip() for x in mc_spec.split("^")]
    if len(esc_split) > 2:
        raise ScrSetupError(
            f"cannot have more than one '^' in match chain specification of '{arg}'"
        )
    if len(esc_split) == 1:
        return parse_simple_mc_range(ctx, mc_spec, arg)
    lhs, rhs = esc_split
    if lhs == "":
        exclude = parse_simple_mc_range(ctx, rhs, arg)
        include: Iterable[match_chain.MatchChain] = itertools.chain(
            ctx.match_chains, [ctx.origin_mc])
    else:
        exclude = parse_simple_mc_range(ctx, rhs, arg)
        chain_count = len(ctx.match_chains)
        include = parse_simple_mc_range(ctx, lhs, arg)
        # hack: parse exclude again so the newly generated chains form include are respected
        if chain_count != len(ctx.match_chains):
            exclude = parse_simple_mc_range(ctx, rhs, arg)
    return ({*include} - {*exclude})


def parse_mc_arg(
    ctx: 'scr_context.ScrContext', argname: str, arg: str,
    support_blank: bool = False, blank_value: str = ""
) -> Optional[tuple[Iterable['match_chain.MatchChain'], str]]:
    if not utils.begins(arg, argname):
        return None
    argname_len = len(argname)
    eq_pos = arg.find("=")
    if eq_pos == -1:
        mc_spec = arg[argname_len:]
        if arg != argname:
            if not MATCH_CHAIN_ARGUMENT_REGEX.match(mc_spec):
                return None
        elif not support_blank:
            raise ScrSetupError("missing equals sign in argument '{arg}'")
        pre_eq_arg = arg
        value = blank_value
    else:
        mc_spec = arg[argname_len: eq_pos]
        if not MATCH_CHAIN_ARGUMENT_REGEX.match(mc_spec):
            return None
        pre_eq_arg = arg[:eq_pos]
        value = arg[eq_pos+1:]
    return parse_mc_range(ctx, mc_spec, pre_eq_arg), value


def parse_mc_arg_as_range(
    ctx: 'scr_context.ScrContext', argname: str, argval: str
) -> list['match_chain.MatchChain']:
    return list(parse_mc_range(ctx, argval, argname))


def apply_mc_arg(
    ctx: 'scr_context.ScrContext', argname: str, config_opt_names: list[str], arg: str,
    value_parse: Callable[[str, str], Any] = lambda x, _arg: x,
    support_blank: bool = False, blank_value: str = ""
) -> bool:
    parse_result = parse_mc_arg(
        ctx, argname, arg, support_blank, blank_value)
    if parse_result is None:
        return False
    mcs, value = parse_result
    value = value_parse(value, arg)
    mcs = list(mcs)
    # so the lowest possible chain generates potential errors
    mcs.sort(key=lambda mc: mc.chain_id if mc.chain_id else float("inf"))
    for mc in mcs:
        prev = mc.try_set_config_option(config_opt_names, value, arg)
        if prev is not None:
            if mc is ctx.origin_mc:
                chainid = str(max(len(ctx.match_chains), 1))
            elif mc is ctx.defaults_mc:
                chainid = ""
            else:
                chainid = str(mc.chain_id)
            raise ScrSetupError(
                f"{argname}{chainid} specified twice in: "
                + f"'{prev}' and '{arg}'"
            )

    return True


def get_arg_val(arg: str) -> str:
    return arg[arg.find("=") + 1:]


def parse_bool_arg(v: str, arg: str, blank_val: bool = True) -> bool:
    v = v.strip().lower()
    if v == "" and blank_val is not None:
        return blank_val

    if v in input_sequences.YES_INDICATING_STRINGS.matching:
        return True
    if v in input_sequences.NO_INDICATING_STRINGS.matching:
        return False
    raise ScrSetupError(f"cannot parse '{v}' as a boolean in '{arg}'")


def parse_int_arg(v: str, arg: str) -> int:
    try:
        return int(v)
    except ValueError:
        raise ScrSetupError(f"cannot parse '{v}' as an integer in '{arg}'")


def parse_non_negative_float_arg(v: str, arg: str) -> float:
    try:
        f = float(v)
    except ValueError:
        raise ScrSetupError(f"cannot parse '{v}' as an number in '{arg}'")
    if f < 0:
        raise ScrSetupError(f"negative number '{v}' not allowed for '{arg}'")
    return f


def parse_encoding_arg(v: str, arg: str) -> str:
    if not verify_encoding(v):
        raise ScrSetupError(f"unknown encoding in '{arg}'")
    return v


def select_variant(val: str, variants_dict: dict[str, T], default: Optional[T] = None) -> Optional[T]:
    val = val.strip().lower()
    if val == "":
        return default
    if val in variants_dict:
        return variants_dict[val]
    match = None
    for k, v in variants_dict.items():
        if utils.begins(k, val):
            if match is not None:
                return None  # we have two conflicting matches
            match = v
    return match


def parse_variant_arg(val: str, variants_dict: dict[str, T], arg: str, default: Optional[T] = None) -> T:
    res = select_variant(val, variants_dict, default)
    if res is None:
        raise ScrSetupError(
            f"illegal argument '{arg}', valid options for "
            + f"{arg[:len(arg)-len(val)-1]} are: "
            + f"{', '.join(sorted(variants_dict.keys()))}"
        )
    return res


def verify_encoding(encoding: str) -> bool:
    try:
        "!".encode(encoding=encoding)
        return True
    except UnicodeEncodeError:
        return False


def apply_doc_arg(
    ctx: 'scr_context.ScrContext', argname: str, doctype: 'DocumentType', arg: str
) -> bool:
    parse_result = parse_mc_arg(ctx, argname, arg)
    if parse_result is None:
        return False
    mcs, path = parse_result
    mcs = list(mcs)
    if mcs == [ctx.defaults_mc]:
        extend_chains_above = len(ctx.match_chains)
        mcs = list(ctx.match_chains)
    elif ctx.origin_mc in mcs:
        mcs.remove(ctx.origin_mc)
        extend_chains_above = len(ctx.match_chains)
    else:
        extend_chains_above = None
    path, path_parsed = scr.normalize_link(
        ctx,
        None,
        doctype.url_handling_type(),
        None,
        None,
        path,
        urllib.parse.urlparse(path)
    )
    doc = document.Document(
        doctype,
        path,
        None,
        mcs,
        extend_chains_above,
        path_parsed=path_parsed
    )
    ctx.docs.append(doc)
    return True


def parse_plain_arg(
    optname: str, arg: str,
    value_parse: Callable[[str, str], Any] = lambda x, _arg: x,
    support_blank: bool = False,
    blank_val: str = ""
) -> Optional[Any]:
    if not utils.begins(arg, optname):
        return None
    if len(optname) == len(arg):
        if support_blank:
            val = blank_val
        else:
            raise ScrSetupError(
                f"missing '=' and value for option '{optname}'"
            )
    else:
        nc = arg[len(optname):]
        if MATCH_CHAIN_ARGUMENT_REGEX.match(nc):
            raise ScrSetupError(
                "option '{optname}' does not support match chain specification"
            )
        if nc[0] != "=":
            return None
        val = get_arg_val(arg)
    return value_parse(val, arg)


def apply_ctx_arg(
    ctx: 'scr_context.ScrContext', optname: str, argname: str, arg: str,
    value_parse: Callable[[str, str], Any] = lambda x, _arg: x,
    support_blank: bool = False,
    blank_val: str = ""
) -> bool:
    val = parse_plain_arg(optname, arg, value_parse, support_blank, blank_val)
    if val is None:
        return False
    if ctx.has_custom_value([argname]):
        raise ScrSetupError(f"error: {argname} specified twice")
    ctx.try_set_config_option([argname], val, arg)
    return True


def parse_selenium_variants(value: str, arg: str) -> Optional['SeleniumVariant']:
    return parse_variant_arg(value, selenium_variants_dict, arg)


def print_version() -> None:
    print(f"{SCRIPT_NAME} {VERSION}")


def parse_args(ctx: 'scr_context.ScrContext', args: Iterable[str]) -> None:
    for arg in args:
        if match_traditional_cli_arg(arg, "help", {"-h", "--help"}):
            help()
            ctx.special_args_occured = True
            continue
        if match_traditional_cli_arg(arg, "version", {"-v", "--version"}):
            print_version()
            ctx.special_args_occured = True
            continue

        driver_install = parse_plain_arg(
            "selinstall", arg, parse_selenium_variants
        )
        if driver_install is not None:
            selenium_driver_download.install_selenium_driver(
                ctx, driver_install, False
            )
            ctx.special_args_occured = True
            continue

        driver_update = parse_plain_arg(
            "selupdate", arg, parse_selenium_variants
        )
        if driver_update is not None:
            selenium_driver_download.install_selenium_driver(
                ctx, driver_update, True
            )
            ctx.special_args_occured = True
            continue

        driver_uninstall = parse_plain_arg(
            "seluninstall", arg, parse_selenium_variants
        )
        if driver_uninstall is not None:
            selenium_driver_download.uninstall_selenium_driver(
                ctx, driver_uninstall
            )
            ctx.special_args_occured = True
            continue

        # content args
        if apply_mc_arg(ctx, "cx", ["loc_content", "xpath"], arg):
            continue
        if apply_mc_arg(ctx, "cr", ["loc_content", "regex"], arg):
            continue
        if apply_mc_arg(ctx, "cf", ["loc_content", "format"], arg):
            continue
        if apply_mc_arg(ctx, "cjs", ["loc_content", "js_script"], arg):
            continue
        if apply_mc_arg(ctx, "cmm", ["loc_content", "multimatch"], arg, parse_bool_arg, True):
            continue
        if apply_mc_arg(ctx, "cin", ["loc_content", "interactive"], arg, parse_bool_arg, True):
            continue
        if apply_mc_arg(ctx, "cfc", ["loc_content"], arg, lambda v, arg: parse_mc_arg_as_range(ctx, arg, v)):
            continue

        if apply_mc_arg(ctx, "cimin", ["cimin"], arg, parse_int_arg):
            continue
        if apply_mc_arg(ctx, "cimax", ["cimax"], arg, parse_int_arg):
            continue
        if apply_mc_arg(ctx, "cicont", ["ci_continuous"], arg, parse_bool_arg, True):
            continue

        if apply_mc_arg(ctx, "cff", ["content_forward_format"], arg):
            continue
        if apply_mc_arg(ctx, "cpf", ["content_print_format"], arg):
            continue
        if apply_mc_arg(ctx, "cwf", ["content_write_format"], arg):
            continue
        if apply_mc_arg(ctx, "csf", ["content_save_format"], arg):
            continue
        if apply_mc_arg(ctx, "csin", ["save_path_interactive"], arg, parse_bool_arg, True):
            continue

        if apply_mc_arg(ctx, "cienc", ["content_input_encoding"], arg, parse_encoding_arg):
            continue
        if apply_mc_arg(ctx, "cfienc", ["content_forced_input_encoding"], arg, parse_encoding_arg):
            continue

        if apply_mc_arg(ctx, "cl", ["content_raw"], arg, lambda v, arg: not parse_bool_arg(v, arg), True): continue
        if apply_mc_arg(ctx, "cesc", ["content_escape_sequence"], arg):
            continue

        # label args
        if apply_mc_arg(ctx, "lx", ["loc_label", "xpath"], arg):
            continue
        if apply_mc_arg(ctx, "lr", ["loc_label", "regex"], arg):
            continue
        if apply_mc_arg(ctx, "lf", ["loc_label", "format"], arg):
            continue
        if apply_mc_arg(ctx, "ljs", ["loc_label", "js_script"], arg):
            continue
        if apply_mc_arg(ctx, "lmm", ["loc_label", "multimatch"], arg, parse_bool_arg, True):
            continue
        if apply_mc_arg(ctx, "lin", ["loc_label", "interactive"], arg, parse_bool_arg, True):
            continue
        if apply_mc_arg(ctx, "las", ["allow_slashes_in_labels"], arg, parse_bool_arg, True):
            continue
        if apply_mc_arg(ctx, "lic", ["labels_inside_content"], arg, parse_bool_arg, True):
            continue
        if apply_mc_arg(ctx, "lam", ["label_allow_missing"], arg, parse_bool_arg, True):
            continue
        if apply_mc_arg(ctx, "ldf", ["label_default_format"], arg, parse_bool_arg, True):
            continue
        if apply_mc_arg(ctx, "fdf", ["filename_default_format"], arg, parse_bool_arg, True):
            continue

        # document args
        if apply_mc_arg(ctx, "dx", ["loc_document", "xpath"], arg):
            continue
        if apply_mc_arg(ctx, "dr", ["loc_document", "regex"], arg):
            continue
        if apply_mc_arg(ctx, "df", ["loc_document", "format"], arg):
            continue
        if apply_mc_arg(ctx, "djs", ["loc_document", "js_script"], arg):
            continue
        if apply_mc_arg(ctx, "doc", ["document_output_chains"], arg, lambda v, arg: parse_mc_arg_as_range(ctx, arg, v)):
            continue
        if apply_mc_arg(ctx, "dmm", ["loc_document", "multimatch"], arg, parse_bool_arg, True):
            continue
        if apply_mc_arg(ctx, "din", ["loc_document", "interactive"], arg, parse_bool_arg, True):
            continue

        if apply_mc_arg(ctx, "dimin", ["dimin"], arg, parse_int_arg):
            continue
        if apply_mc_arg(ctx, "dimax", ["dimax"], arg, parse_int_arg):
            continue

        if apply_mc_arg(ctx, "owf", ["overwrite_files"], arg, parse_bool_arg, True):
            continue

        if apply_mc_arg(ctx, "denc", ["default_document_encoding"], arg, parse_encoding_arg):
            continue
        if apply_mc_arg(ctx, "dfenc", ["forced_document_encoding"], arg, parse_encoding_arg):
            continue

        if apply_mc_arg(ctx, "dsch", ["default_document_scheme"], arg):
            continue
        if apply_mc_arg(ctx, "dpsch", ["prefer_parent_document_scheme"], arg):
            continue
        if apply_mc_arg(ctx, "dfsch", ["forced_document_scheme"], arg):
            continue

        if apply_mc_arg(
            ctx, "selstrat", ["selenium_strategy"], arg,
            lambda v, arg: parse_variant_arg(v, selenium_strats_dict, arg)
        ): continue
        if apply_mc_arg(
            ctx, "seldl", ["selenium_download_strategy"], arg,
            lambda v, arg: parse_variant_arg(
                v, selenium_download_strategies_dict, arg)
        ): continue

        # misc args
        if apply_doc_arg(ctx, "url", DocumentType.URL, arg):
            continue
        if apply_doc_arg(ctx, "rfile", DocumentType.RFILE, arg):
            continue
        if apply_doc_arg(ctx, "file", DocumentType.FILE, arg):
            continue

        if apply_ctx_arg(ctx, "cookiefile", "cookie_file", arg):
            continue

        if apply_ctx_arg(
            ctx, "sel", "selenium_variant", arg,
            lambda v, arg: parse_variant_arg(
                v, selenium_variants_dict, arg, SeleniumVariant.FIREFOX
            ),
            True
        ): continue
        if apply_ctx_arg(ctx, "selh", "selenium_headless", arg, parse_bool_arg, True):
            continue
        if apply_ctx_arg(ctx, "selkeep", "selenium_keep_alive", arg, parse_bool_arg, True):
            continue
        if apply_ctx_arg(ctx, "tbdir", "tor_browser_dir", arg):
            continue
        if apply_ctx_arg(ctx, "bfs", "documents_bfs", arg, parse_bool_arg, True):
            continue
        if apply_ctx_arg(ctx, "ua", "user_agent", arg):
            continue
        if apply_ctx_arg(ctx, "uar", "user_agent_random", arg, parse_bool_arg, True):
            continue
        if apply_ctx_arg(ctx, "v", "verbosity", arg, lambda v, arg: parse_variant_arg(v, verbosities_dict, arg)):
            continue

        if apply_ctx_arg(ctx, "repl", "repl", arg,  parse_bool_arg, True):
            continue

        if apply_ctx_arg(ctx, "mt", "max_download_threads", arg,  parse_int_arg):
            continue

        if apply_ctx_arg(ctx, "--repl", "repl", arg,  parse_bool_arg, True):
            continue

        if apply_ctx_arg(ctx, "exit", "exit", arg,  parse_bool_arg, True):
            continue

        if apply_ctx_arg(ctx, "timeout", "request_timeout_seconds", arg,  parse_non_negative_float_arg):
            continue

        raise ScrSetupError(f"unrecognized option: '{arg}'")



from typing import Optional, Union, cast
import math
import datetime
from collections import deque
import sys
import urllib
import os
from . import download_job, utils

DOWNLOAD_STATUS_LOG_ELEMENTS_MIN = 5
DOWNLOAD_STATUS_LOG_ELEMENTS_MAX = 50
DOWNLOAD_STATUS_LOG_MAX_AGE = 10
DOWNLOAD_STATUS_NAME_LENGTH = 80
DOWNLOAD_STATUS_BAR_LENGTH = 30
DOWNLOAD_STATUS_REFRESH_INTERVAL = 0.2
DOWNLOAD_STATUS_KEEP_FINISHED = True


def get_byte_size_string(size: Union[int, float]) -> tuple[str, str]:
    if size < 2**10:
        if type(size) is int:
            return f"{size}", "B"
        return f"{size:.2f}", "B"
    units = ["K", "M", "G", "T", "P", "E", "Z", "Y"]
    unit = int(math.log(size, 1024))
    if unit >= len(units):
        unit = len(units)
    return f"{float(size)/2**(10 * unit):.2f}", f"{units[unit - 1]}iB"


def get_timespan_string(ts: float) -> tuple[str, str]:
    if round(ts * 10) < 600:
        return f"{ts:.1f}", "s"
    if round(ts) == 60:
        return "01:00", "m"
    if ts < 3600:
        return f"{int(ts / 60):02}:{round(ts % 60):02}", "m"
    return f"{int(ts / 3600):02}:{int((ts % 3600) / 60):02}:{int(ts % 60):02}", "h"


def lpad(string: str, tgt_len: int) -> str:
    return " " * (tgt_len - len(string)) + string


def rpad(string: str, tgt_len: int) -> str:
    return string + " " * (tgt_len - len(string))


class DownloadStatusReport:
    name: str
    expected_size: Optional[int] = None
    downloaded_size: int = 0
    download_begin_time: datetime.datetime
    download_end_time: Optional[datetime.datetime] = None
    updates: deque[tuple[datetime.datetime, int]]
    download_finished: bool = False
    download_manager: 'download_job.DownloadManager'

    def __init__(self, download_manager: 'download_job.DownloadManager') -> None:
        self.updates = deque()
        self.download_manager = download_manager

    def gen_display_name(
        self,
        url: Optional[urllib.parse.ParseResult],
        filename: Optional[str],
        save_path: Optional[str]
    ) -> None:
        if save_path:
            if len(save_path) < DOWNLOAD_STATUS_NAME_LENGTH:
                self.name = save_path
                return
            self.name = os.path.basename(save_path)
        elif filename:
            self.name = filename
        elif url is not None:
            self.name = url.geturl()
        else:
            self.name = "<unnamed download>"
        self.name = utils.truncate(
            self.name, DOWNLOAD_STATUS_NAME_LENGTH
        )

    def submit_update(self, received_filesize: int) -> None:
        time = datetime.datetime.now()
        with self.download_manager.status_report_lock:
            self.downloaded_size += received_filesize
            self.updates.append((time, self.downloaded_size))
            drop_elem = False
            if len(self.updates) > DOWNLOAD_STATUS_LOG_ELEMENTS_MIN:
                if len(self.updates) > DOWNLOAD_STATUS_LOG_ELEMENTS_MAX:
                    drop_elem = True
                else:
                    age = (time - self.updates[0][0]).total_seconds()
                    if age > DOWNLOAD_STATUS_LOG_MAX_AGE:
                        drop_elem = True
            if drop_elem:
                self.updates.popleft()

    def enqueue(self) -> None:
        with self.download_manager.status_report_lock:
            self.download_manager.download_status_reports.append(self)
        self.download_begin_time = datetime.datetime.now()

    def finished(self) -> None:
        self.download_end_time = datetime.datetime.now()
        with self.download_manager.status_report_lock:
            self.download_finished = True


class StatusReportLine:
    name: str
    expected_size: Optional[int]
    downloaded_size: int
    speed_calculatable: bool
    download_begin: datetime.datetime
    download_end: Optional[datetime.datetime]
    speed_frame_time_begin: datetime.datetime
    speed_frame_time_end: datetime.datetime
    speed_frame_size_begin: int
    speed_frame_size_end: int
    star_pos: int = 1
    star_dir: int = 1
    last_line_length: int = 0
    finished: bool = False

    total_time_str: str
    total_time_u_str: str
    bar_str: str
    downloaded_size_str: str
    downloaded_size_u_str: str
    expected_size_str: str
    expected_size_u_str: str
    speed_str: str
    speed_u_str: str
    eta_str: str
    eta_u_str: str


class ProgressReportManager:
    finished_report_lines: list[StatusReportLine]
    finished_report_lines_max_length: int = 0
    newly_finished_report_lines: list[StatusReportLine]
    report_lines: list[StatusReportLine]
    prev_report_line_count: int
    prev_terminal_column_count: int = 0

    total_time_lm: int = 0
    total_time_u_lm: int = 0
    downloaded_size_lm: int = 0
    downloaded_size_u_lm: int = 0
    expected_size_lm: int = 0
    expected_size_u_lm: int = 0
    eta_lm: int = 0
    eta_u_lm: int = 0
    speed_lm: int = 0
    speed_u_lm: int = 0
    lms_changed: bool = False

    def __init__(self) -> None:
        self.report_lines = []
        self.finished_report_lines = []
        self.newly_finished_report_lines = []
        self.prev_report_line_count = 0

    def load_status(self, download_manager: 'download_job.DownloadManager') -> None:
        with download_manager.status_report_lock:
            self._load_status_report_lines(
                download_manager.download_status_reports
            )

    def updates_remaining(self) -> bool:
        return (
            len(self.report_lines) > 0
            or
            len(self.newly_finished_report_lines) > 0
        )

    def _load_status_report_lines(self, dsr_list: list[DownloadStatusReport]) -> None:
        if (not DOWNLOAD_STATUS_KEEP_FINISHED) and len(dsr_list) > len(self.report_lines):
            # when we have more reports than report lines,
            # we remove the oldest finished report
            # if none are finished, we get more report lines
            i = 0
            while i < len(dsr_list):
                if dsr_list[i].download_finished:
                    del dsr_list[i]
                    if len(dsr_list) == len(self.report_lines):
                        break
                else:
                    i += 1
        for i in range(len(dsr_list) - len(self.report_lines)):
            self.report_lines.append(StatusReportLine())
        for i in range(len(self.report_lines)):
            rl = self.report_lines[i]
            dsr = dsr_list[i]
            rl.name = dsr.name
            rl.expected_size = dsr.expected_size
            rl.downloaded_size = dsr.downloaded_size
            rl.download_begin = dsr.download_begin_time
            rl.download_end = dsr.download_end_time
            rl.finished = dsr.download_finished
            if not len(dsr.updates):
                rl.speed_calculatable = False
            elif len(dsr.updates) == 1:
                rl.speed_calculatable = True
                rl.speed_frame_time_begin = dsr.download_begin_time
                rl.speed_frame_size_begin = 0
                rl.speed_frame_time_end = dsr.updates[0][0]
                rl.speed_frame_size_end = dsr.updates[0][1]
            else:
                rl.speed_calculatable = True
                rl.speed_frame_time_begin = dsr.updates[0][0]
                rl.speed_frame_size_begin = dsr.updates[0][1]
                rl.speed_frame_time_end = dsr.updates[-1][0]
                rl.speed_frame_size_end = dsr.updates[-1][1]
        if DOWNLOAD_STATUS_KEEP_FINISHED:
            i = 0
            while i < len(dsr_list):
                if dsr_list[i].download_finished:
                    self.newly_finished_report_lines.append(self.report_lines[i])
                    del dsr_list[i]
                    del self.report_lines[i]
                else:
                    i += 1

    def _stringify_status_report_lines(self, report_lines: list[StatusReportLine]) -> None:
        now = datetime.datetime.now()
        for rl in report_lines:
            if rl.expected_size and rl.expected_size >= rl.downloaded_size:
                frac = float(rl.downloaded_size) / rl.expected_size
                filled = int(frac * (DOWNLOAD_STATUS_BAR_LENGTH - 1))
                empty = DOWNLOAD_STATUS_BAR_LENGTH - filled - 1
                tip = ">" if rl.downloaded_size != rl.expected_size else "="
                rl.bar_str = "[" + "=" * filled + tip + " " * empty + "]"
            elif rl.finished:
                rl.bar_str = "[" + "*" * DOWNLOAD_STATUS_BAR_LENGTH + "]"
            else:
                left = rl.star_pos - 1
                right = DOWNLOAD_STATUS_BAR_LENGTH - 3 - left
                rl.bar_str = "[" + " " * left + "***" + " " * right + "]"
                if rl.star_pos == DOWNLOAD_STATUS_BAR_LENGTH - 2:
                    rl.star_dir = -1
                elif rl.star_pos == 1:
                    rl.star_dir = 1
                rl.star_pos += rl.star_dir
            if rl.finished:
                rl.expected_size = rl.downloaded_size
            rl.downloaded_size_str, rl.downloaded_size_u_str = (
                get_byte_size_string(rl.downloaded_size)
            )
            if rl.expected_size:
                rl.expected_size_str, rl.expected_size_u_str = (
                    get_byte_size_string(rl.expected_size)
                )
            else:
                rl.expected_size_str, rl.expected_size_u_str = "???", "B"

            if rl.finished:
                assert rl.download_end
                rl.speed_frame_size_begin = 0
                rl.speed_frame_time_begin = rl.download_begin
                rl.speed_frame_size_end = rl.downloaded_size
                rl.speed_frame_time_end = rl.download_end
                rl.speed_calculatable = True
            else:
                rl.download_end = now
            if rl.speed_calculatable:
                duration = (
                    (rl.speed_frame_time_end -
                        rl.speed_frame_time_begin).total_seconds()
                )
                handled_size = rl.speed_frame_size_end - rl.speed_frame_size_begin
                if duration < sys.float_info.epsilon:
                    rl.speed_str, rl.speed_u_str = "???", " "
                    rl.eta_str, rl.eta_u_str = "???", " "
                else:
                    if handled_size == 0:
                        speed = 0.0
                        rl.eta_str, rl.eta_u_str = "???", " "
                    else:
                        speed = float(handled_size) / duration
                        if rl.expected_size and rl.expected_size > rl.downloaded_size:
                            rl.eta_str, rl.eta_u_str = get_timespan_string(
                                (rl.expected_size - rl.downloaded_size) / speed
                            )
                        elif rl.finished:
                            rl.eta_str, rl.eta_u_str = "---", "-"
                        else:
                            rl.eta_str, rl.eta_u_str = "???", " "
                    rl.speed_str, rl.speed_u_str = get_byte_size_string(speed)
                    rl.speed_u_str += "/s"
            else:
                rl.speed_frame_time_end = now
                rl.eta_str, rl.eta_u_str = "???", " "
                rl.speed_str, rl.speed_u_str = "???", ""

            rl.total_time_str, rl.total_time_u_str = get_timespan_string(
                (rl.download_end - rl.download_begin).total_seconds()
            )

    def _update_field_len_max(
        self, report_lines: list[StatusReportLine], field_name: str
    ) -> None:
        rls_lm = max(map(
            lambda rl: len(rl.__dict__[field_name + "_str"]), report_lines
        ), default=0)
        if rls_lm > cast(int, self.__dict__.get(field_name + "_lm", 0)):
            self.__dict__[field_name + "_lm"] = rls_lm
            self.lms_changed = True

    def _append_status_report_line_strings(
        self, report_lines: list[StatusReportLine], report: list[str]
    ) -> None:

        self._update_field_len_max(report_lines, "total_time")
        self._update_field_len_max(report_lines, "total_time_u")
        self._update_field_len_max(report_lines, "downloaded_size")
        self._update_field_len_max(report_lines, "downloaded_size_u")
        self._update_field_len_max(report_lines, "expected_size")
        self._update_field_len_max(report_lines, "expected_size_u")
        self._update_field_len_max(report_lines, "eta")
        self._update_field_len_max(report_lines, "eta_u")
        self._update_field_len_max(report_lines, "speed")
        self._update_field_len_max(report_lines, "speed_u")

        for rl in report_lines:
            line = ""

            line += lpad(rl.total_time_str, self.total_time_lm) + " "
            line += rpad(rl.total_time_u_str, self.total_time_u_lm) + " "

            line += lpad(rl.speed_str, self.speed_lm) + " "
            line += rpad(rl.speed_u_str, self.speed_u_lm) + " "

            line += rl.bar_str + " "

            line += lpad(rl.downloaded_size_str, self.downloaded_size_lm) + " "
            line += rpad(rl.downloaded_size_u_str, self.downloaded_size_u_lm)
            line += " / "
            line += lpad(rl.expected_size_str, self.expected_size_lm) + " "
            line += rpad(rl.expected_size_u_str, self.expected_size_u_lm) + " "

            line += "eta "
            line += lpad(rl.eta_str, self.eta_lm)
            line += " " + rpad(rl.eta_u_str, self.eta_u_lm) + " "

            # if everybody is done, we don't need to pad for this anymore
            line += rl.name

            if len(line) < rl.last_line_length:
                lll = len(line)
                # fill with spaces to clear previous line
                line += " " * (rl.last_line_length - lll)
                rl.last_line_length = lll
            else:
                rl.last_line_length = len(line)
            report.append(line)

    def print_status_report(self) -> None:
        max_cols = os.get_terminal_size().columns
        if max_cols < 5:
            # don't bother
            return
        # even if the size just grew we are doing this since the terminal
        # might have messed up our output
        update_finished = (max_cols != self.prev_terminal_column_count)
        report_lines = self.newly_finished_report_lines + self.report_lines
        self._stringify_status_report_lines(report_lines)
        report_line_strings: list[str] = []
        self._append_status_report_line_strings(
            self.newly_finished_report_lines,
            report_line_strings
        )
        max_len_nfrls = max((len(rls) for rls in report_line_strings), default=0)
        self.finished_report_lines_max_length = max(
            max_len_nfrls,
            self.finished_report_lines_max_length
        )
        self._append_status_report_line_strings(
            self.report_lines,
            report_line_strings
        )
        if self.lms_changed:
            self._stringify_status_report_lines(self.finished_report_lines)
            update_finished = True
            self.lms_changed = False

        report_lines_to_rewrite = self.prev_report_line_count
        if update_finished:
            finished_rls: list[str] = []
            self._append_status_report_line_strings(
                self.finished_report_lines,
                finished_rls
            )
            finished_rls.extend(report_line_strings)
            report_line_strings = finished_rls
            report_lines_to_rewrite += len(self.finished_report_lines)
        report = ""
        if report_lines_to_rewrite:
            report += f"\x1B[{report_lines_to_rewrite}F"
        for rls in report_line_strings:
            rls_len = len(rls)
            if rls_len < max_cols:
                report += rls + " " * (max_cols - len(rls)) + "\n"
            else:
                report += rls[0:max_cols-3] + "...\n"
        sys.stdout.write(report)
        self.finished_report_lines.extend(self.newly_finished_report_lines)
        self.newly_finished_report_lines.clear()
        self.prev_report_line_count = len(self.report_lines)
        self.prev_terminal_column_count = max_cols



from typing import Iterable

from .definitions import (T)


def prefixes(str: str) -> set[str]:
    return set(str[:i] for i in range(len(str), 0, -1))


def set_join(*args: Iterable['T']) -> set['T']:
    res: set[T] = set()
    for s in args:
        res.update(s)
    return res


class OptionIndicatingStrings:
    representative: str
    matching: set[str]

    def __init__(self, representative: str, *args: set[str]) -> None:
        self.representative = representative
        if args:
            self.matching = set_join(*args)
        else:
            self.matching = prefixes(representative)


YES_INDICATING_STRINGS = OptionIndicatingStrings(
    "yes",
    prefixes("yes"), prefixes("true"), {"1", "+"}
)
NO_INDICATING_STRINGS = OptionIndicatingStrings(
    "no", prefixes("no"), prefixes("false"), {"0", "-"}
)
SKIP_INDICATING_STRINGS = OptionIndicatingStrings("skip")
CHAIN_SKIP_INDICATING_STRINGS = OptionIndicatingStrings("chainskip")
EDIT_INDICATING_STRINGS = OptionIndicatingStrings("edit")
DOC_SKIP_INDICATING_STRINGS = OptionIndicatingStrings("docskip")
INSPECT_INDICATING_STRINGS = OptionIndicatingStrings("inspect")
ACCEPT_CHAIN_INDICATING_STRINGS = OptionIndicatingStrings("acceptchain")



from typing import Optional, Union
from .definitions import (
    SeleniumStrategy, SeleniumDownloadStrategy, DocumentType,
    DEFAULT_ESCAPE_SEQUENCE, FALLBACK_DOCUMENT_SCHEME)
from .config_data_class import ConfigDataClass
from . import locator, content_match, scr_context, document


class MatchChain(ConfigDataClass):
    # config members
    # this is a config member so it is copied on apply_defaults
    ctx: 'scr_context.ScrContext'
    content_escape_sequence: str = DEFAULT_ESCAPE_SEQUENCE
    cimin: int = 1
    cimax: Union[int, float] = float("inf")
    ci_continuous: bool = False
    content_save_format: Optional[str] = None
    content_print_format: Optional[str] = None
    content_write_format: Optional[str] = None
    content_forward_chains: list['MatchChain'] = []
    content_raw: bool = True
    content_input_encoding: str = "utf-8"
    content_forced_input_encoding: Optional[str] = None
    save_path_interactive: bool = False

    label_default_format: Optional[str] = None
    filename_default_format: Optional[str] = None
    labels_inside_content: bool = False
    label_allow_missing: bool = False
    allow_slashes_in_labels: bool = False
    overwrite_files: bool = True

    dimin: int = 1
    dimax: Union[int, float] = float("inf")
    default_document_encoding: str = "utf-8"
    forced_document_encoding: Optional[str] = None

    default_document_scheme: str = FALLBACK_DOCUMENT_SCHEME
    prefer_parent_document_scheme: bool = True
    forced_document_scheme: Optional[str] = None

    selenium_strategy: SeleniumStrategy = SeleniumStrategy.PLAIN
    selenium_download_strategy: SeleniumDownloadStrategy = SeleniumDownloadStrategy.EXTERNAL

    document_output_chains: list['MatchChain']
    __annotations__: dict[str, type]
    _config_slots_: list[str] = (
        ConfigDataClass._previous_annotations_as_config_slots(
            __annotations__, [])
    )

    # subconfig members
    loc_content: 'locator.Locator'
    loc_label: 'locator.Locator'
    loc_document: 'locator.Locator'

    _subconfig_slots_ = ['loc_content', 'loc_label', 'loc_document']

    # non config members
    chain_id: int
    di: int
    ci: int
    js_executed: bool = False
    has_xpath_matching: bool = False
    has_label_matching: bool = False
    has_content_xpaths: bool = False
    # TODO: this should include if this is the target of any doc=...
    has_document_matching: bool = False
    has_content_matching: bool = False
    parses_documents = False  # used for the document as content optimization
    has_interactive_matching: bool = False
    need_content: bool = False
    need_label: bool = False
    need_filename: bool = False
    need_output_multipass: bool = False
    content_matches: list['content_match.ContentMatch']
    document_matches: list['document.Document']
    handled_content_matches: set['content_match.ContentMatch']
    handled_document_matches: set['document.Document']
    satisfied: bool = True
    labels_none_for_n: int = 0

    def __init__(self, ctx: 'scr_context.ScrContext', chain_id: int, blank: bool = False) -> None:
        super().__init__(blank)

        self.ctx = ctx
        self.chain_id = chain_id
        self.document_output_chains = []

        self.loc_content = locator.Locator("content", blank)
        self.loc_label = locator.Locator("label", blank)
        self.loc_document = locator.Locator("document", blank)

        self.content_matches = []
        self.document_matches = []
        self.handled_content_matches = set()
        self.handled_document_matches = set()

    def gen_dummy_document(self) -> 'document.Document':
        d = document.Document(
            DocumentType.FILE, "", None,
            locator_match=self.loc_document.gen_dummy_locator_match()
        )
        d.encoding = ""
        return d

    def gen_dummy_content_match(self) -> 'content_match.ContentMatch':
        clm = self.loc_content.gen_dummy_locator_match()
        if self.has_label_matching:
            llm = self.loc_label.gen_dummy_locator_match()
        elif self.label_default_format:
            llm = locator.LocatorMatch()
            llm.fres = ""
        else:
            llm = None

        dcm = content_match.ContentMatch(
            clm, llm, self, self.gen_dummy_document())
        if self.loc_content.multimatch:
            dcm.ci = 0
        if self.has_document_matching:
            dcm.di = 0
        return dcm

    def accepts_content_matches(self) -> bool:
        return self.di <= self.dimax

    def need_document_matches(self, current_di_used: int) -> bool:
        return (
            self.has_document_matching
            and self.di <= (self.dimax - (1 if current_di_used else 0))
        )

    def need_content_matches(self) -> bool:
        assert self.ci is not None
        assert self.di is not None
        return self.has_content_matching and self.ci <= self.cimax and self.di <= self.dimax

    def is_valid_label(self, label: str) -> bool:
        if self.allow_slashes_in_labels:
            return True
        if "/" in label or "\\" in label:
            return False
        return True



from .definitions import (
    DocumentType, URL_FILENAME_MAX_LEN, Verbosity, InteractiveResult,
    SeleniumDownloadStrategy, ScrFetchError, SeleniumVariant, DEFAULT_CWF
)
from .input_sequences import (
    YES_INDICATING_STRINGS, NO_INDICATING_STRINGS, EDIT_INDICATING_STRINGS,
    CHAIN_SKIP_INDICATING_STRINGS, DOC_SKIP_INDICATING_STRINGS
)
from typing import Optional, BinaryIO, Union, cast, Iterator
import os
import urllib
import time
import sys
from tbselenium.tbdriver import TorBrowserDriver
from selenium.common.exceptions import WebDriverException as SeleniumWebDriverException
from selenium.webdriver.remote.webdriver import WebDriver as SeleniumWebDriver
import binascii
import requests
from . import (
    progress_report, content_match, scr, utils, scr_context, selenium_setup
)
from collections import OrderedDict
import threading
import concurrent.futures
from abc import ABC, abstractmethod
from enum import Enum

DEFAULT_MAX_PRINT_BUFFER_CAPACITY = 2**20 * 100  # 100 MiB
DEFAULT_RESPONSE_BUFFER_SIZE = 32768


class ContentFormat(Enum):
    STRING = 0,
    BYTES = 1,
    STREAM = 2,
    FILE = 3,
    TEMP_FILE = 4,
    UNNEEDED = 5,


class PrintOutputManager:
    printing_buffers: OrderedDict[int, list[bytes]]
    finished_queues: set[int]
    lock: threading.Lock
    size_blocked: threading.Condition
    size_limit: int
    dl_ids: int = 0
    active_id: int = 0
    main_thread_id: Optional[int] = None

    def __init__(self, max_buffer_size: int = DEFAULT_MAX_PRINT_BUFFER_CAPACITY) -> None:
        self.lock = threading.Lock()
        self.printing_buffers = OrderedDict()
        self.finished_queues = set()
        self.size_limit = max_buffer_size
        self.size_blocked = threading.Condition(self.lock)

    def reset(self) -> None:
        self.active_id = 0
        self.dl_ids = 0
        self.main_thread_id = self.request_print_access()

    def main_thread_done(self) -> None:
        if self.main_thread_id is not None:
            self.declare_done(self.main_thread_id)
            self.main_thread_id = None

    def print(self, id: int, buffer: bytes) -> None:
        is_active = False
        with self.lock:
            while True:
                if id == self.active_id:
                    is_active = True
                    stored_buffers = self.printing_buffers.pop(id, [])
                    self.size_limit += sum(
                        map(lambda b: len(b), stored_buffers)
                    )
                    break
                elif self.size_limit > len(buffer):
                    self.size_limit -= len(buffer)
                    self.printing_buffers[id].append(buffer)
                    break
                self.size_blocked.wait()
        if is_active:
            for b in stored_buffers:
                sys.stdout.buffer.write(b)
            sys.stdout.buffer.write(buffer)
            if(stored_buffers):
                self.size_blocked.notifyAll()

    def request_print_access(self) -> int:
        with self.lock:
            id = self.dl_ids
            self.dl_ids += 1
            if id != self.active_id:
                self.printing_buffers[id] = []
        return id

    def try_reaquire_main_thread_print_access(self) -> bool:
        with self.lock:
            if self.dl_ids != self.active_id:
                return False
            self.main_thread_id = self.dl_ids
            self.dl_ids += 1
        return True

    def declare_done(self, id: int) -> None:
        new_active_id = None
        buffers_to_print: list[list[bytes]] = []
        with self.lock:
            if self.active_id != id:
                self.finished_queues.add(id)
                return

            new_active_id = self.active_id + 1
            while new_active_id in self.finished_queues:
                self.finished_queues.remove(new_active_id)
                buffers_to_print.append(
                    self.printing_buffers.pop(new_active_id)
                )
                new_active_id += 1
        while True:
            for bl in buffers_to_print:
                for b in bl:
                    sys.stdout.buffer.write(b)
            # after we printed and reacquire the lock, the job
            # that we want to give the active_id token to
            # might have finished already, in which case we have to print him too
            buffers_to_print.clear()
            with self.lock:
                self.active_id = new_active_id
                if new_active_id not in self.finished_queues:
                    new_active_id = None
                    break
                while True:
                    self.finished_queues.remove(new_active_id)
                    buffers_to_print.append(
                        self.printing_buffers.pop(new_active_id)
                    )
                    new_active_id += 1
                    if new_active_id not in self.finished_queues:
                        break
            if new_active_id is None:
                break

    def flush(self, id: int) -> None:
        with self.lock:
            if not id != self.active_id:
                return
        sys.stdout.flush()


class PrintOutputStream:
    pom: PrintOutputManager
    id: int

    def __init__(self, pom: PrintOutputManager) -> None:
        self.pom = pom
        self.id = pom.request_print_access()

    def write(self, buffer: bytes) -> int:
        self.pom.print(self.id, buffer)
        return len(buffer)

    def flush(self) -> None:
        self.pom.flush(self.id)

    def close(self) -> None:
        self.pom.declare_done(self.id)


class MinimalInputStream(ABC):
    @abstractmethod
    def read(self, size: Optional[int]) -> bytes:
        pass

    @abstractmethod
    def close(self) -> None:
        pass

    @abstractmethod
    def __enter__(self) -> None:
        pass

    @abstractmethod
    def __exit__(self) -> None:
        pass


class ResponseStreamWrapper(MinimalInputStream):
    _bytes_buffer: bytearray
    _request_response: requests.models.Response
    _iterator: Iterator[bytes]
    _pos: int = 0

    def __init__(
        self, request_response: requests.models.Response,
        buffer_size: int = DEFAULT_RESPONSE_BUFFER_SIZE
    ) -> None:
        self._bytes_buffer = bytearray()
        self._request_response = request_response
        self._iterator = self._request_response.iter_content(buffer_size)

    def read(self, size: Optional[int] = None) -> bytes:
        if size is None:
            goal_position = float("inf")
        else:
            goal_position = self._pos + size

        loaded_until = self._pos + len(self._bytes_buffer)
        while loaded_until < goal_position:
            try:
                buf = next(self._iterator)
            except StopIteration:
                goal_position = loaded_until
                break
            loaded_until += len(buf)
            if self._bytes_buffer:
                self._bytes_buffer.extend(buf)
            else:
                self._bytes_buffer = bytearray(buf)
        if loaded_until <= goal_position:
            self._pos = loaded_until
            res = self._bytes_buffer
            self._bytes_buffer = bytearray()
            return res
        assert type(goal_position) is int
        buf_pos = goal_position - self._pos
        self._pos = goal_position
        res = self._bytes_buffer[0:buf_pos]
        self._bytes_buffer = self._bytes_buffer[buf_pos:]
        return res

    def close(self) -> None:
        self._request_response.close()

    def __enter__(self) -> None:
        pass

    def __exit__(self) -> None:
        self.close()


class DownloadJob:
    save_file: Optional[BinaryIO] = None
    temp_file: Optional[BinaryIO] = None
    temp_file_path: Optional[str] = None
    multipass_file: Optional[BinaryIO] = None
    print_stream: Optional[PrintOutputStream] = None
    content_stream: Union[BinaryIO, MinimalInputStream, None] = None
    content: Union[str, bytes, BinaryIO, MinimalInputStream, None] = None
    content_format: Optional[ContentFormat] = None
    filename: Optional[str] = None
    status_report: Optional['progress_report.DownloadStatusReport'] = None

    cm: 'content_match.ContentMatch'
    save_path: Optional[str] = None
    context: str
    output_formatters: list['scr.OutputFormatter']

    def __init__(self, cm: content_match.ContentMatch) -> None:
        self.cm = cm
        self.context = (
            f"{utils.truncate(self.cm.doc.path)}{scr.get_ci_di_context(self.cm)}"
        )
        self.output_formatters = []

    def requires_download(self) -> bool:
        return self.cm.mc.need_content and not self.cm.mc.content_raw

    def setup_print_stream(self, pom: 'PrintOutputManager') -> None:
        if self.cm.mc.content_print_format is not None:
            self.print_stream = PrintOutputStream(pom)

    def request_status_report(self, download_manager: 'DownloadManager') -> None:
        self.status_report = progress_report.DownloadStatusReport(
            download_manager
        )

    def gen_fallback_filename(self, dont_use_url: bool = False) -> bool:
        if self.filename is not None or not self.cm.mc.need_filename:
            return True
        if not dont_use_url:
            path = cast(urllib.parse.ParseResult, self.cm.url_parsed).path
            self.filename = scr.sanitize_filename(urllib.parse.unquote(path))
            if self.filename is not None and len(self.filename) < URL_FILENAME_MAX_LEN:
                return True
        try:
            self.filename = scr.gen_final_content_format(
                cast(str, self.cm.mc.filename_default_format), self.cm, None
            ).decode("utf-8", errors="surrogateescape")
            return True
        except UnicodeDecodeError:
            scr.log(
                self.cm.mc.ctx, Verbosity.ERROR,
                f"{self.cm.doc.path}{scr.get_ci_di_context(self.cm)}: "
                + "generated default filename not valid utf-8"
            )
            return False

    def handle_save_path(self) -> 'InteractiveResult':
        if self.save_path is not None:
            # this was already done during for interactive filename determination
            return InteractiveResult.ACCEPT
        cm = self.cm
        if not cm.mc.content_save_format:
            return InteractiveResult.ACCEPT
        if cm.llm and not cm.mc.is_valid_label(cm.llm.result):
            scr.log(cm.mc.ctx, Verbosity.WARN,
                    f"matched label '{cm.llm.result}' would contain a slash, skipping this content from: {cm.doc.path}"
                    )
            save_path = None
        if cm.mc.need_filename:
            if not self.fetch_content():
                return InteractiveResult.ERROR
        save_path_bytes = scr.gen_final_content_format(
            cm.mc.content_save_format, cm, self.filename
        )
        try:
            save_path = save_path_bytes.decode(
                "utf-8", errors="surrogateescape"
            )
        except UnicodeDecodeError:
            scr.log(
                cm.mc.ctx, Verbosity.ERROR,
                f"{cm.doc.path}{scr.get_ci_di_context(cm)}: generated save path is not valid utf-8"
            )
            save_path = None
        while True:
            if save_path and not os.path.exists(os.path.dirname(os.path.abspath(save_path))):
                scr.log(cm.mc.ctx, Verbosity.ERROR,
                        f"{cm.doc.path}{scr.get_ci_di_context(cm)}: directory of generated save path does not exist"
                        )
                save_path = None
            if not save_path and not cm.mc.save_path_interactive:
                return InteractiveResult.ERROR
            if not cm.mc.save_path_interactive:
                break
            if save_path:
                res = scr.prompt(
                    f'{cm.doc.path}{scr.get_ci_di_context(cm)}: accept save path "{save_path}" [Yes/no/edit/chainskip/docskip]? ',
                    [
                        (InteractiveResult.ACCEPT, YES_INDICATING_STRINGS),
                        (InteractiveResult.REJECT, NO_INDICATING_STRINGS),
                        (InteractiveResult.EDIT, EDIT_INDICATING_STRINGS),
                        (InteractiveResult.SKIP_CHAIN,
                         CHAIN_SKIP_INDICATING_STRINGS),
                        (InteractiveResult.SKIP_DOC, DOC_SKIP_INDICATING_STRINGS)
                    ],
                    InteractiveResult.ACCEPT
                )
                if res == InteractiveResult.ACCEPT:
                    break
                if res != InteractiveResult.EDIT:
                    return res
            save_path = input("enter new save path: ")
        if save_path is None:
            return InteractiveResult.REJECT
        self.save_path = save_path
        return InteractiveResult.ACCEPT

    def selenium_download_from_local_file(self) -> bool:
        path = self.cm.clm.result
        self.content_format = ContentFormat.FILE
        if cast(urllib.parse.ParseResult, self.cm.url_parsed).scheme == "file":
            offs = len("file:")
            for i in range(2):
                if path[offs] == "/":
                    offs += 1
            path = path[offs:]
        self.content = path
        self.filename = os.path.basename(path)
        return True

    def selenium_download_external(self) -> bool:
        proxies = None
        if self.cm.mc.ctx.selenium_variant == SeleniumVariant.TORBROWSER:
            tbdriver = cast(TorBrowserDriver, self.cm.mc.ctx.selenium_driver)
            proxies = {
                "http": f"socks5h://localhost:{tbdriver.socks_port}",
                "https": f"socks5h://localhost:{tbdriver.socks_port}",
                "data": None
            }
        try:
            try:
                req = scr.request_raw(
                    self.cm.mc.ctx, self.cm.clm.result, cast(
                        urllib.parse.ParseResult, self.cm.url_parsed),
                    selenium_setup.load_selenium_cookies(self.cm.mc.ctx),
                    proxies=proxies, stream=True
                )
                self.content = ResponseStreamWrapper(req)
                self.content_format = ContentFormat.STREAM
                self.filename = scr.request_try_get_filename(req)
                if self.status_report:
                    self.status_report.expected_size = (
                        scr.request_try_get_filesize(req)
                    )
                return True
            except requests.exceptions.RequestException as ex:
                raise scr.request_exception_to_scr_fetch_error(ex)
        except ScrFetchError as ex:
            scr.log(
                self.cm.mc.ctx, Verbosity.ERROR,
                f"{utils.truncate(self.cm.doc.path)}{scr.get_ci_di_context(self.cm)}: "
                + f"failed to download '{utils.truncate(self.cm.clm.result)}': {str(ex)}"
            )
            return False

    def selenium_download_internal(self) -> bool:
        doc_url_str = selenium_setup.selenium_get_url(self.cm.mc.ctx)
        if doc_url_str is None:
            return False
        doc_url = urllib.parse.urlparse(doc_url_str)

        if doc_url.netloc != cast(urllib.parse.ParseResult, self.cm.url_parsed).netloc:
            scr.log(
                self.cm.mc.ctx, Verbosity.ERROR,
                f"{self.cm.clm.result}{scr.get_ci_di_context(self.cm)}: "
                + "failed to download: seldl=internal does not work across origins"
            )
            return False

        tmp_path, tmp_filename = scr.gen_dl_temp_name(self.cm.mc.ctx, None)
        script_source = """
            const url = arguments[0];
            const filename = arguments[1];
            const a = document.createElement('a');
            a.href = url;
            a.download = filename;
            document.body.appendChild(a);
            a.click();
            document.body.removeChild(a);
        """
        try:
            selenium_setup.selenium_exec_script(self.cm.mc.ctx, script_source,
                                                self.cm.clm.result, tmp_filename)
        except SeleniumWebDriverException as ex:
            if selenium_setup.selenium_has_died(self.cm.mc.ctx):
                selenium_setup.report_selenium_died(self.cm.mc.ctx)
            else:
                scr.log(
                    self.cm.mc.ctx, Verbosity.ERROR,
                    f"{self.cm.clm.result}{scr.get_ci_di_context(self.cm)}: "
                    + f"selenium download failed: {str(ex)}"
                )
            return False
        i = 0
        while True:
            if os.path.exists(tmp_path):
                time.sleep(0.1)
                break
            if i < 10:
                time.sleep(0.01)
            else:
                time.sleep(0.1)
                if i > 15:
                    i = 10
                    if selenium_setup.selenium_has_died(self.cm.mc.ctx):
                        return False

            i += 1
        self.content = tmp_path
        self.content_format = ContentFormat.TEMP_FILE
        # TODO: maybe support filenames here ?
        return True

    def selenium_download_fetch(self) -> bool:
        script_source = """
            const url = arguments[0];
            var content_disposition = null;
            return (async () => {
                return await fetch(url, {
                    method: 'GET',
                })
                .then(res => {
                    content_disposition = res.headers.get(
                        'Content-Disposition');
                    return res.blob();
                })
                .then((blob, cd) => new Promise((resolve, reject) => {
                    const reader = new FileReader();
                    reader.readAsDataURL(blob);
                    reader.onload = () => (resolve(reader.result.substr(reader.result.indexOf(',') + 1)), cd);
                    reader.onerror = error => reject(error);
                }))
                .then(result => {
                    return {
                        "ok": result,
                        "content_disposition": content_disposition,
                    };
                })
                .catch(ex => {
                    return {
                        "error": ex.message
                    };
                });
            })();
        """
        err = None
        driver = cast(SeleniumWebDriver, self.cm.mc.ctx.selenium_driver)
        try:
            doc_url = driver.current_url
            res = selenium_setup.selenium_exec_script(
                self.cm.mc.ctx, script_source, self.cm.clm.result)
        except SeleniumWebDriverException as ex:
            if selenium_setup.selenium_has_died(self.cm.mc.ctx):
                selenium_setup.report_selenium_died(self.cm.mc.ctx)
                return False
            err = str(ex)
        if "error" in res:
            err = res["error"]
        if err is not None:
            cors_warn = ""
            if urllib.parse.urlparse(doc_url).netloc != urllib.parse.urlparse(self.cm.clm.result).netloc:
                cors_warn = " (potential CORS issue)"
            scr.log(
                self.cm.mc.ctx, Verbosity.ERROR,
                f"{utils.truncate(self.cm.doc.path)}{scr.get_ci_di_context(self.cm)}: "
                + f"selenium download of '{self.cm.clm.result}' failed{cors_warn}: {err}"
            )
            return False
        self.content = binascii.a2b_base64(res["ok"])
        if self.status_report:
            self.status_report.expected_size = len(self.content)
        self.filename = scr.try_get_filename_from_content_disposition(
            res.get("content_disposition", "")
        )
        self.content_format = ContentFormat.BYTES
        return True

    def selenium_download(self) -> bool:
        if (
            self.cm.doc.document_type == DocumentType.FILE
            and cast(urllib.parse.ParseResult, self.cm.url_parsed).scheme in ["", "file"]
        ):
            if not self.selenium_download_from_local_file():
                return False
        elif self.cm.mc.selenium_download_strategy == SeleniumDownloadStrategy.EXTERNAL:
            if not self.selenium_download_external():
                return False
        elif self.cm.mc.selenium_download_strategy == SeleniumDownloadStrategy.INTERNAL:
            if not self.selenium_download_internal():
                return False
        else:
            assert self.cm.mc.selenium_download_strategy == SeleniumDownloadStrategy.FETCH
            if not self.selenium_download_fetch():
                return False
        if not self.gen_fallback_filename():
            return False
        return True

    def fetch_content(self) -> bool:
        if self.content_format is not None:
            # this was already done during filename determination
            return True
        if self.cm.mc.content_raw:
            self.content = self.cm.clm.result
            self.content_format = ContentFormat.STRING
            if not self.gen_fallback_filename():
                return False
        else:
            if not self.cm.mc.need_content:
                self.content_format = ContentFormat.UNNEEDED
                # even if we don't need the content, somebody might want the filename
                # although we don't use Content-Disposition deduction here,
                # which is debatable because it is somewhat inconsistent
                # especially for data urls, this is quite terrible
                if not self.gen_fallback_filename():
                    return False
            else:
                if self.cm.mc.ctx.selenium_variant.enabled():
                    if not self.selenium_download():
                        return False
                else:
                    data = scr.try_read_data_url(self.cm)
                    if data is not None:
                        self.content = data
                        self.content_format = ContentFormat.BYTES
                        if self.status_report:
                            self.status_report.expected_size = len(data)
                        if not self.gen_fallback_filename(dont_use_url=True):
                            return False
                    elif self.cm.doc.document_type.derived_type() is DocumentType.FILE:
                        self.content = self.cm.clm.result
                        self.content_format = ContentFormat.FILE
                        if self.status_report:
                            try:
                                self.status_report.expected_size = os.path.getsize(
                                    self.content)
                            except IOError:
                                pass
                        if not self.gen_fallback_filename():
                            return False
                    else:
                        try:
                            res = scr.request_raw(
                                self.cm.mc.ctx, self.cm.clm.result,
                                cast(
                                    urllib.parse.ParseResult,
                                    self.cm.url_parsed
                                ),
                                stream=True
                            )
                            self.content = ResponseStreamWrapper(res)
                            self.filename = scr.request_try_get_filename(res)
                            if self.status_report:
                                self.status_report.expected_size = (
                                    scr.request_try_get_filesize(res)
                                )
                            self.content_format = ContentFormat.STREAM
                            if not self.gen_fallback_filename():
                                return False
                        except requests.exceptions.RequestException as ex:
                            fe = scr.request_exception_to_scr_fetch_error(ex)
                            scr.log(self.cm.mc.ctx, Verbosity.ERROR,
                                    f"{self.context}: failed to download '{utils.truncate(self.cm.clm.result)}': {str(fe)}")
                            return False
        return True

    def setup_save_file(self) -> bool:
        if not self.save_path:
            return True
        try:
            use_as_multipass = (
                self.cm.mc.need_output_multipass
                and self.multipass_file is None
                and self.cm.mc.content_write_format == DEFAULT_CWF
            )
            save_file = cast(BinaryIO, open(
                self.save_path,
                ("w" if self.cm.mc.overwrite_files else "x")
                + "b"
                + ("+" if use_as_multipass else "")
            ))
            if use_as_multipass:
                self.multipass_file = save_file
        except FileExistsError:
            scr.log(self.cm.mc.ctx, Verbosity.ERROR,
                    f"{self.context}: file already exists: {self.save_path}")
            return False
        except OSError as ex:
            scr.log(
                self.cm.mc.ctx, Verbosity.ERROR,
                f"{self.context}: failed to write to file '{self.save_path}': {str(ex)}"
            )
            return False

        self.output_formatters.append(scr.OutputFormatter(
            cast(str, self.cm.mc.content_write_format),
            self.cm, save_file, self.content, self.filename
        ))
        return True

    def setup_content_file(self) -> bool:
        if self.content_format not in [ContentFormat.FILE, ContentFormat.TEMP_FILE]:
            return True
        assert type(self.content) is str
        try:
            self.content_stream = cast(BinaryIO, scr.fetch_file(
                self.cm.mc.ctx, self.content, stream=True)
            )
        except ScrFetchError as ex:
            scr.log(self.cm.mc.ctx, Verbosity.ERROR,
                    f"{self.context}: failed to open file '{utils.truncate(self.content)}': {str(ex)}")
            return False
        if self.content_format == ContentFormat.TEMP_FILE:
            self.temp_file_path = self.content
        self.content = self.content_stream
        if self.cm.mc.need_output_multipass:
            self.multipass_file = self.content_stream
        return True

    def setup_print_output(self) -> bool:
        if self.cm.mc.content_print_format is None:
            return True
        if self.print_stream is not None:
            stream: Union[PrintOutputStream, BinaryIO] = self.print_stream
        else:
            stream = sys.stdout.buffer
        self.output_formatters.append(scr.OutputFormatter(
            self.cm.mc.content_print_format, self.cm,
            stream, self.content, self.filename
        ))
        return True

    def check_abort(self) -> None:
        if self.cm.mc.ctx.abort:
            raise InterruptedError

    def run_job(self) -> bool:
        if self.status_report:
            self.status_report.gen_display_name(
                self.cm.url_parsed, self.filename, self.save_path
            )
            self.status_report.enqueue()
        success = False
        try:
            if self.handle_save_path() != InteractiveResult.ACCEPT:
                return False
            if not self.fetch_content():
                return False

            self.check_abort()
            self.content_stream: Union[BinaryIO, MinimalInputStream, None] = (
                cast(Union[BinaryIO, MinimalInputStream], self.content)
                if self.content_format == ContentFormat.STREAM
                else None
            )

            if not self.setup_content_file():
                return False
            if not self.setup_save_file():
                return False
            if self.status_report:
                # try to generate a better name now that we have more information
                self.status_report.gen_display_name(
                    self.cm.url_parsed, self.filename, self.save_path
                )
            if not self.setup_print_output():
                return False
            self.check_abort()

            if self.content_stream is None:
                if self.status_report and self.content:
                    self.status_report.submit_update(
                        len(cast(Union[str, bytes], self.content)))
                for of in self.output_formatters:
                    res = of.advance()
                    assert not res
                    self.check_abort()
                success = True
                return True

            if self.cm.mc.need_output_multipass and self.multipass_file is None:
                try:
                    self.temp_file_path, _filename = scr.gen_dl_temp_name(
                        self.cm.mc.ctx, self.save_path)
                    self.temp_file = open(self.temp_file_path, "xb+")
                except IOError:
                    return False
                self.multipass_file = self.temp_file
                self.check_abort()

            if self.content_stream is not None:
                while True:
                    buf = self.content_stream.read(
                        DEFAULT_RESPONSE_BUFFER_SIZE)
                    self.check_abort()
                    if buf is None:
                        continue
                    if self.status_report:
                        self.status_report.submit_update(len(buf))
                    advance_output_formatters(self.output_formatters, buf)
                    if self.temp_file:
                        self.temp_file.write(buf)
                    if len(buf) < DEFAULT_RESPONSE_BUFFER_SIZE:
                        if self.content_stream is not self.multipass_file:
                            self.content_stream.close()
                            self.content_stream = None
                        break

            if self.multipass_file:
                while self.output_formatters:
                    self.multipass_file.seek(0)
                    while True:
                        buf = self.multipass_file.read(
                            DEFAULT_RESPONSE_BUFFER_SIZE)
                        self.check_abort()
                        advance_output_formatters(self.output_formatters, buf)
                        if len(buf) < DEFAULT_RESPONSE_BUFFER_SIZE:
                            break
            success = True
            return True
        except InterruptedError:
            return False
        finally:
            if self.status_report:
                self.status_report.finished()
            if self.print_stream is not None:
                self.print_stream.close()
            if self.content_stream is not None:
                self.content_stream.close()
            if self.temp_file is not None:
                self.temp_file.close()
            if self.temp_file_path is not None:
                os.remove(self.temp_file_path)
            if self.save_file is not None:
                self.save_file.close()
            path = self.cm.clm.result
            if self.requires_download():
                scr.log(self.cm.mc.ctx, Verbosity.DEBUG,
                        f"finished downloading {path}" if success else f"failed to download {path}"
                        )


class DownloadManager:
    ctx: 'scr_context.ScrContext'
    max_threads: int
    pending_jobs: set[concurrent.futures.Future[bool]]
    pom: PrintOutputManager
    executor: concurrent.futures.ThreadPoolExecutor
    status_report_lock: threading.Lock
    download_status_reports: list['progress_report.DownloadStatusReport']
    enable_status_reports: bool

    def __init__(self, ctx: 'scr_context.ScrContext', max_threads: int, enable_status_reports: bool) -> None:
        self.ctx = ctx
        self.pending_jobs = set()
        self.executor = concurrent.futures.ThreadPoolExecutor(
            max_workers=max_threads
        )
        self.pom = PrintOutputManager()
        self.status_report_lock = threading.Lock()
        self.download_status_reports = []
        self.enable_status_reports = enable_status_reports

    def submit(self, dj: DownloadJob) -> None:
        scr.log(
            self.ctx, Verbosity.DEBUG,
            f"enqueuing download for {dj.cm.clm.result}"
        )
        dj.setup_print_stream(self.pom)
        if self.enable_status_reports:
            dj.request_status_report(self)
        self.pending_jobs.add(self.executor.submit(dj.run_job))

    def wait_until_jobs_done(self) -> None:
        if not self.pending_jobs:
            return
        may_print = False
        if self.pom:
            may_print = self.pom.try_reaquire_main_thread_print_access()
        if not self.enable_status_reports or not may_print:
            results = concurrent.futures.wait(self.pending_jobs)
            for x in results.done:
                x.result()
            self.pending_jobs.clear()
        self.pom.request_print_access()
        prm = progress_report.ProgressReportManager()
        while True:
            results = concurrent.futures.wait(
                self.pending_jobs,
                timeout=0 if not prm.prev_report_line_count
                else progress_report.DOWNLOAD_STATUS_REFRESH_INTERVAL
            )
            for x in results.done:
                x.result()
            self.pending_jobs = results.not_done
            prm.load_status(self)
            if not prm.updates_remaining():
                if not self.pending_jobs:
                    # this happens when we got main thread print access
                    # but everybody is already done and never downloaded anything
                    # we don't want any progress reports here
                    break
                continue
            prm.print_status_report()
            if not self.pending_jobs:
                break

    def terminate(self, cancel_running: bool = False) -> None:
        try:
            if not cancel_running:
                cancel_running = True
                self.wait_until_jobs_done()
                cancel_running = False
        finally:
            if cancel_running:
                self.ctx.abort = True
            self.executor.shutdown(wait=True, cancel_futures=cancel_running)


def advance_output_formatters(output_formatters: list['scr.OutputFormatter'], buf: Optional[bytes]) -> None:
    i = 0
    while i < len(output_formatters):
        if output_formatters[i].advance(DEFAULT_RESPONSE_BUFFER_SIZE, buf):
            i += 1
        else:
            del output_formatters[i]
